/* -*- Mode: C; tab-width: 8; c-basic-offset: 4 -*- */
/* vim:set sw=4 ts=8: */
/*
 * This file is part of the MUSE Instrument Pipeline
 * Copyright (C) 2008-2014 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

/*********************************************************************
 * E.S.O. - VLT project
 *
 * "@(#)"
 *
 * who       when        what
 * --------  ----------  ----------------------------------------------
 * apecontal 2013-04-16  improved overscan detection within slices
 * apecontal 2013-03-01  fixed bug computing overscan median value when 
 *                       the binning in y direction is greater than 1
 * apecontal 2013-02-19  added usage of overscan pixels to estimate bias
 * apecontal 2012-10-30  fixed bug regarding handling of binned mask
 * apecontal 2012-10-24  fixed bug regarding handling of
 *                       cpl_bivector_interpolate_linear output according to
 *                       version number
 * gzins     2012-10-23  passed slices limits as argument
 * apecontal 2012-09-21  read slices limits into mask instead or computing them
 * apecontal 2011-11-24  bug fix in slices ref (updated wr_row computation)
 * gzins     2011-03-09  suppressed muclipm_make_image_zero function
 * gzins     2011-03-05  re-added muclipm_make_image_zero
 * apecontal 2011-03-03  added documentation
 *                       improved error handling
 *                       moved defines to header file, and renamed them
 * apecontal 2010-02-08  added capability of masks with unfixed width 
 * apecontal 2009-12-15  created
 */

/**
 * @defgroup muclipm_make_image Fast Image Reconstruction 
 *
 * Function used to derive a rough estimate of the field of view observed 
 * by MUSE (using spectrographic data as input). This is a fast and dirty 
 * reconstruction mode and by the way less accurate than processing
 * the data using the pipeline. 
 *
 * @par Synopsis:
 * @code
 *   #include "muclipm_make_image.h"
 * @endcode
 */

/*-----------------------------------------------------------------------------
    Includes
 -----------------------------------------------------------------------------*/

#include "muclipm_make_image.h"
#include "muclipm_priv_error.h"

#include <string.h>
 
/*-----------------------------------------------------------------------------
    Implementation
 -----------------------------------------------------------------------------*/
/**@{*/

/*----------------------------------------------------------------------------*/
/**
 * @brief   Evaluates per quadrant overscan median values and subtract them from
 *          input image
 *
 * @param   raw_img      the raw image to correct from overscan offset
 *
 * @return  CPL error code
 * 
 * @par Introduction
 * This function evaluates the median value of the vertical overscan strip of 
 * each CCD quadrant and correct each quadrant of the input image with corres-
 * ponding value
 * 
 * @par Product Creation:
 * The input @a cpl_image is updated
 * 
 * @error{return CPL_ERROR_NULL_INPUT, raw_img or mask are NULL}
 * @error{return CPL_ERROR_INCOMPATIBLE_INPUT, raw_img has wrong dimensions}
 * 
 * @code
 *   #include "muclipm_make_image.h"
 * @endcode
 */
/*----------------------------------------------------------------------------*/
cpl_error_code  muclipm_bias_subtract_using_overscan(cpl_image *raw_img)
{
    long raw_image_nx = cpl_image_get_size_x(raw_img); 
    long raw_image_ny = cpl_image_get_size_y(raw_img); 
    long i, j, q;

    int  err=0, shiftx, shifty;
    double median = 0;

    double Qd[4][4]; /* Limits of CCD quadrants */
    double Ov[4][4]; /* Limits of overscan strips */

    if (raw_img == NULL) {
	return(CPL_ERROR_INCOMPATIBLE_INPUT);
    }

    if (raw_image_nx < 64 || raw_image_ny < 64) {
	return(CPL_ERROR_INCOMPATIBLE_INPUT);
    }

    /* compute CCD quadrant limits */
    for (q=0; q<4; q++) 
    {
	if (q % 2 == 0)
            shiftx = 0;
        else
	    shiftx = 1;

	if (q > 1)
            shifty = 1;
        else
	    shifty = 0;

        Qd[q][0] = 0.5*shiftx*raw_image_nx + 1;
        Qd[q][1] = Qd[q][0] -1 +0.5*raw_image_nx;
        Qd[q][2] = 0.5*shifty*raw_image_ny + 1;
        Qd[q][3] = Qd[q][2] -1 +0.5*raw_image_ny;
    } 

    /* Compute overscan strip limits per quadrant (ignore the 4 pixels all around the strip) */
    for (q=0; q<4; q++)  
    {
	if (q % 2 == 0)
	    Ov[q][0] = Qd[q][1] -32 +4;
	else
	    Ov[q][0] = Qd[q][0] +4;

	Ov[q][1] = Ov[q][0] +32 - 2*4;

	Ov[q][2] = Qd[q][2]+4;
	Ov[q][3] = Qd[q][3]-4;
    }

    /* for each CCD quadrant, estimate median of the overscan strip, and subtract from image */
    for (q=0; q<4; q++) 
    {
	/* compute median in overscan strip */
	median = cpl_image_get_median_window(raw_img,Ov[q][0],Ov[q][2],Ov[q][1],Ov[q][3]);	

	for (j=Qd[q][2];j<=Qd[q][3];j++)
	    for (i=Qd[q][0];i<=Qd[q][1];i++)
		cpl_image_set(raw_img,i,j,cpl_image_get(raw_img,i,j,&err) -median);
    }

    return(CPL_ERROR_NONE);
}

/*----------------------------------------------------------------------------*/
/**
 * @brief   Collapses spectra over the spectral direction
 *
 * @param   raw_img      the raw science image
 * @param   offset       the bias image
 * @param   flat         the flat-field image
 * @param   mask         the mask image
 * @param   slice_xstart the start position of the slices along X
 * @param   slice_xend   the end position of the slices along X
 * @param   out_img      the output image
 *
 * @return  CPL error code
 * 
 * @par Introduction
 * This function rebuilds from the spectrographic data the image of the field of
 * view belonging from one IFU. 
 * Any binning may be used in the spectral direction while a binning in the
 * spatial direction will result in strange behaviour.
 * 
 * @par Product Creation:
 * The returned product is a @a cpl_image
 * 
 * @note
 * If flat calibration frame is not present (NULL pointer) then
 * corresponding corrections are not applied. If the bias calibration frame is
 * not passed (NULL pointer), then the offset values (1 per CCD quadrant) are 
 * evaluated using the overscan strips and applied as a correction. No error 
 * message raised in both cases.
 * 
 * @error{return CPL_ERROR_NULL_INPUT, raw_img or mask are NULL}
 * @error{return CPL_ERROR_INCOMPATIBLE_INPUT, raw_img has wrong dimensions}
 * 
 * @code
 *   #include "muclipm_make_image.h"
 * @endcode
 */
/*----------------------------------------------------------------------------*/
cpl_error_code  muclipm_make_image(const cpl_image    *raw_img,
                                   const cpl_image    *offset,
                                   const cpl_image    *flat,
                                   const cpl_image    *mask,
                                   const long         *slice_xstart,
                                   const long         *slice_xend,
                                   cpl_image         **out_img)
{
    /* Local variables */
    long i, j, k, l, m, n, min, imin, imax, outvec_size;
    long raw_image_nx = cpl_image_get_size_x(raw_img); 
    long raw_image_ny = cpl_image_get_size_y(raw_img); 
    long wr_col, wr_row, noslice;
    int  err=0, xoverscan=0;
    float cpl_version;
    double val=0;
    double *invec_x_data, *outvec_x_data;
    double *invec_y_data, *outvec_y_data;

    cpl_bivector *invec, *outvec;
    cpl_vector *invec_x, *invec_y;
    cpl_image *tmp_img, *collapsed_img;

    /* Slit position per sub-slicer  (1/4th of the slicer) */
    int slit_position[MUCLIPM_NB_SLICES] = 
        {9, 8, 1, 10, 7, 2, 11, 6, 3, 12, 5, 4};

    MUCLIPM_TRY
    {
        MUCLIPM_TRY_CHECK(raw_img != NULL, CPL_ERROR_NULL_INPUT,
                          "raw image", "is mandatory");

        /* Check e.g. for same matrix dimensions */
        if (raw_image_ny <= 0)
        {
            MUCLIPM_TRY_EXIT_WITH_ERROR_MSG
                (CPL_ERROR_INCOMPATIBLE_INPUT, "raw image",
                 "should not be empty");
        }
	/* init output image as blank image (in case the process fails before completion */
	err = cpl_image_multiply_scalar(*out_img,(double)0.0);
        if (err)
        {
            MUCLIPM_TRY_EXIT_WITH_ERROR_MSG
                (cpl_error_get_code(), "output image", "could not set values");
        }

        tmp_img = cpl_image_duplicate(raw_img);
        if (tmp_img == NULL)
        {
            MUCLIPM_TRY_EXIT_WITH_ERROR_MSG
                (cpl_error_get_code(), "temporary image", "could not create");
        }

        /* Subtract the bias if offset pointer is valid  - if not use overscan pixels to estimate it */
        if (offset != NULL) 
        {
            err = cpl_image_subtract(tmp_img, offset);
            if (err != CPL_ERROR_NONE)
            {
                cpl_image_delete(tmp_img);
                return(err);
            }
        }
	else
        {
	    muclipm_bias_subtract_using_overscan(tmp_img);
        }

        /* Divide by the flat if flat pointer is valid */
        if (flat != NULL) 
        {
            err = cpl_image_divide(tmp_img, flat);
            if (err != CPL_ERROR_NONE)
            {
                cpl_image_delete(tmp_img);
                return(err);
            }
        }

        /* Apply the mask (i.e. keep  nly relevant data) */
        MUCLIPM_TRY_CHECK(mask != NULL, CPL_ERROR_NULL_INPUT,
                          "mask image", "is mandatory");

        err = cpl_image_multiply(tmp_img, mask); 
        if (err != CPL_ERROR_NONE)
        {
            cpl_image_delete(tmp_img);
            return(err);
        }

        /* Collapse image over rows to keep only .... */
        collapsed_img = cpl_image_collapse_create(tmp_img,0);

        if (collapsed_img == NULL) 
        {
            cpl_image_delete(tmp_img);
            MUCLIPM_TRY_EXIT_WITH_ERROR_MSG
                (cpl_error_get_code(), "collapsed image", "could not create");
        }

        cpl_image_delete(tmp_img);

        /* Retrieve cpl version to cope with cpl_bivector_interpolate_linear bug
         * fixed in version 5.2 which doesn't set the first value
         */              
        cpl_version = cpl_version_get_major() + 0.1*cpl_version_get_minor();

        /* Create outvec bivector */
        if (cpl_version >= 5.2) {
            outvec = cpl_bivector_new(MUCLIPM_NB_SPEC_PER_SLICE);
            imin = 0;
            imax = MUCLIPM_NB_SPEC_PER_SLICE;
        }
        else
        {
            outvec = cpl_bivector_new(MUCLIPM_NB_SPEC_PER_SLICE + 1);
            imin = 1;
            imax = MUCLIPM_NB_SPEC_PER_SLICE + 1;
        }

        if (outvec == NULL) 
        {
            MUCLIPM_TRY_EXIT_WITH_ERROR_MSG
                (cpl_error_get_code(), "bivector outvec", "could not create");
        }

        outvec_x_data = cpl_bivector_get_x_data(outvec);
        outvec_y_data = cpl_bivector_get_y_data(outvec);

        outvec_x_data[0] = 1.;
        for (m = imin; m < imax ; m++) 
        {
            outvec_x_data[m] = (double)(m + 1.0);
        }

        /* For each subslicer 1-4*/
        for (k= 1; k <= MUCLIPM_NB_SUBSLICERS; k++)
        {
            /* For each slice 1-12*/
            for (l= 1; l <= MUCLIPM_NB_SLICES; l++)
            {
                noslice = (k-1)*MUCLIPM_NB_SLICES + l-1;

                n = slice_xend[noslice] - slice_xstart[noslice] +1;

                if ((n - 2*MUCLIPM_OVERSCAN) > (MUCLIPM_NB_SPEC_PER_SLICE -2)) 
                {
	            xoverscan = 1;
                    n -= 2*MUCLIPM_OVERSCAN;
                }
		else
		{
	            xoverscan = 0;
		}

                /* Create bivector */
                invec_x = cpl_vector_new(n);
                invec_y = cpl_vector_new(n);
                if (invec_x == NULL || invec_y == NULL) 
                {
                    cpl_image_delete(collapsed_img);
                    MUCLIPM_TRY_EXIT_WITH_ERROR_MSG
                        (cpl_error_get_code(), "vector invecx or invecy",
                         "could not create");
                }

                invec_x_data = cpl_vector_get_data(invec_x);
                invec_y_data = cpl_vector_get_data(invec_y);

                for (i=slice_xstart[noslice], j=0; 
                     (i <= slice_xend[noslice]) && (i < raw_image_nx) ; i++) 
                {

                    val = (double)cpl_image_get(collapsed_img, i, 1, &err);
                    if (err)
                    {
                        val = 0;
                    }

                    /* Store values in a bivector */
                    if (xoverscan != 1)  /* care about slices 
						  crossing overscans window */
                    {
                        invec_x_data[j] = 
                            (double)((j + 1.0) * 
                                     (MUCLIPM_NB_SPEC_PER_SLICE + 1) / n);
                        invec_y_data[j] = val;
                        j++;
                    }
		    else {

			if ((i <= raw_image_nx/2 - MUCLIPM_OVERSCAN) 
			 || (i > raw_image_nx/2 + MUCLIPM_OVERSCAN))
			{
                            invec_x_data[j] = 
                                (double)((j + 1.0) * 
                                     (MUCLIPM_NB_SPEC_PER_SLICE + 1) / n);
                            invec_y_data[j] = val;
                            j++;
			}
		    }
                }

                /* Create bivector */
                invec = cpl_bivector_wrap_vectors(invec_x, invec_y);
                if (invec == NULL)
                {
                    cpl_image_delete(collapsed_img);
                    cpl_bivector_delete(outvec);
                    cpl_bivector_delete(invec);
                    MUCLIPM_TRY_EXIT_WITH_ERROR_MSG
                        (cpl_error_get_code(), "bivector invec", 
                         "could not create");
                }

                outvec_size= cpl_bivector_get_size(outvec);

                if (outvec_size >= n) 
                {
                    /* Only need to copy vector invec in outvec */
                    min = (n < MUCLIPM_NB_SPEC_PER_SLICE) ? 
                        n : MUCLIPM_NB_SPEC_PER_SLICE;

                    for (i=imin, m= 0; i < imax && m < min; i++, m++)
                        outvec_y_data[i] = invec_y_data[m];
                } 
                else
                {
                    /* Interpolate */
                    err = cpl_bivector_interpolate_linear(outvec, invec);
                    if (err)
                    {
                        cpl_image_delete(collapsed_img);
                        cpl_bivector_delete(outvec);
                        cpl_bivector_delete(invec);
                        MUCLIPM_TRY_EXIT_WITH_ERROR_MSG
                            (err, "bivector invec", "could not interpolate");
                    }
                }

                /* Write result */
                wr_row = MUCLIPM_NB_SLICES - slit_position[l - 1] + 1;
                wr_col = k * MUCLIPM_NB_SPEC_PER_SLICE;
                for (i=imin; i< imax ; i++, wr_col--)
                {
                    cpl_image_set(*out_img, wr_col, wr_row,
                                  (float)outvec_y_data[i]);
                }
                cpl_bivector_delete(invec);
            }
        }

        cpl_bivector_delete(outvec);
        cpl_image_delete(collapsed_img);
        cpl_error_reset();
    }
    MUCLIPM_CATCH
    {       
        *out_img = NULL;
    }

    return MUCLIPM_ERROR_GET_NEW_SINCE_TRY();
}

/**@}*/
