/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* vim:set sw=2 sts=2 et cin: */
/*
 * This file is part of the MUSE Instrument Pipeline
 * Copyright (C) 2005-2014 European Southern Observatory
 *           (C) 2000-2002 European Southern Observatory
 *           (C) 2002-2006 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*----------------------------------------------------------------------------*
 *                             Includes                                       *
 *----------------------------------------------------------------------------*/
#include <cpl.h>
#include <math.h>
#include <string.h>

#include "muse_astro.h"

#include "muse_pfits.h"

/*----------------------------------------------------------------------------*
 *                             Debugging Macros                               *
 *         Set these to 1 or higher for (lots of) debugging output            *
 *----------------------------------------------------------------------------*/
#define MUSE_ASTRO_AIRMASS_APPROX 2 /* airmass approximation to use: *
                                     * 0   Young & Irvine (1967)     *
                                     * 1   Young (1994)              *
                                     * 2   Hardie (1962)             */

/*----------------------------------------------------------------------------*/
/**
 * @defgroup muse_astro        Astronomical tools
 *
 * Functions in this group are related to handling of astronomical properties.
 * They are based on pilastroutils/giastroutils taken from the VIMOS and GIRAFFE
 * pipelines (written by Carlo Izzo and Ralf Palsa, (C) 2000-2004 and 2002-2006
 * European Southern Observatory).
 */
/*----------------------------------------------------------------------------*/

/**@{*/

/*----------------------------------------------------------------------------*/
/**
  @private
  @brief   Compute the zenith distance for an observation.
  @param   aHourAngle   Hour angle in radians.
  @param   aDelta       Declination in radians.
  @param   aLatitude    Latitude of the observatory in radians.
  @return  cos(z) on success or 0. on error.

  The function computes the cosine of the zenith distance for an observation
  taken at an angle aHourAngle from the meridian, which can take values in the
  range extending from @f$-\pi@f$ to @f$\pi@f$, and the declination aDelta with
  possible values between @f$-0.5\pi@f$ and @f$0.5\pi@f$. The latitude
  aLatitude of the observing site may take values in the range @f$0@f$ to
  @f$2\pi@f$.
 */
/*----------------------------------------------------------------------------*/
static double
muse_astro_get_zenith_distance(double aHourAngle, double aDelta,
                               double aLatitude)
{
  double p0 = sin(aLatitude) * sin(aDelta),
         p1 = cos(aLatitude) * cos(aDelta),
         z = p0 + cos(aHourAngle) * p1;
  return fabs(z) < FLT_EPSILON ? 0. : z;
}

#if MUSE_ASTRO_AIRMASS_APPROX == 0
/*----------------------------------------------------------------------------*/
/**
  @private
  @brief   Compute the zenith distance for an observation.
  @param   aSecZ   Secans of the zenith distance.
  @return  The function returns the airmass.

  The function uses the approximation given by Young and Irvine (Young A. T.,
  Irvine W. M., 1967, Astron. J. 72, 945) to compute the airmass for a given
  sec(z) aSecZ. This approximation takes into account atmosphere refraction and
  curvature, but is in principle only valid at sea level.
 */
/*----------------------------------------------------------------------------*/
static double
muse_astro_get_airmass_youngirvine(double aSecZ)
{
  return aSecZ * (1. - 0.0012 * (pow(aSecZ, 2) - 1.));
}
#endif /* MUSE_ASTRO_AIRMASS_APPROX == 0 */

#if MUSE_ASTRO_AIRMASS_APPROX == 1
/*----------------------------------------------------------------------------*/
/**
  @private
  @brief   Compute the zenith distance for an observation.
  @param   aCosZt   The cosine of the true zenith distance.
  @return  The function returns the airmass.

  The function uses the approximation given by Young (Young A. T., 1994 ApOpt,
  33, 1108) to compute the relative optical air mass as a function of true,
  rather than refracted, zenith angle which is given in terms of its cosine
  aCosZt.
  It is supposedy more accurate than Young & Irvine (1967) but restrictions are
  not known.
 */
/*----------------------------------------------------------------------------*/
static double
muse_astro_get_airmass_young(double aCosZt)
{
  return (1.002432 * aCosZt*aCosZt + 0.148386 * aCosZt + 0.0096467)
         / (aCosZt*aCosZt*aCosZt + 0.149864 * aCosZt*aCosZt + 0.0102963 * aCosZt
            + 0.000303978);
}
#endif /* MUSE_ASTRO_AIRMASS_APPROX == 1 */

#if MUSE_ASTRO_AIRMASS_APPROX == 2
/*----------------------------------------------------------------------------*/
/**
  @private
  @brief   Compute the zenith distance for an observation.
  @param   aSecZ   The secans of the zenith distance.
  @return  The function returns the airmass.

  The function uses the approximation given by Hardie (Hardie 1962, In:
  "Astronomical Techniques", ed. Hiltner, p. 184) to compute the airmass as a
  function of zenith angle which is given in terms of its secons aSecZ.
  It is supposedy more accurate than Young & Irvine (1967) and usable for zenith
  angles below 85 degrees.
 */
/*----------------------------------------------------------------------------*/
static double
muse_astro_get_airmass_hardie(double aSecZ)
{
  double secm1 = aSecZ - 1;
  return aSecZ - 0.0018167 * secm1 - 0.002875 * secm1*secm1
         - 0.0008083 * secm1*secm1*secm1;
}
#endif /* MUSE_ASTRO_AIRMASS_APPROX == 2 */

/*----------------------------------------------------------------------------*/
/**
  @brief   Compute the effective airmass of an observation.
  @param   aRA         right ascension in degrees
  @param   aDEC        declination in degrees
  @param   aLST        local sidereal time in seconds elapsed since siderial
                       midnight
  @param   aExptime    integration time in seconds
  @param   aLatitude   latitude of the observatory site in degrees
  @return  The computed average airmass or -1 on error.

  The function calculates the average airmass for the line of sight given by the
  right ascension aRA and the declination aDEC. The latitude aLatitude in
  degrees of the observatory site and the local siderial time aLST at
  observation start has to be given, as well as the duration of the observation,
  i.e. the exposure time aExptime. If aExptime is zero then only one value of
  airmass is computed, instead of weighting beginning, middle, and end of
  exposure according to Stetson (Stetson P., 1987, PASP 99, 191).

  This function uses the approximation given by Hardie (Hardie 1962, In:
  "Astronomical Techniques", ed. Hiltner, p. 184) to compute the airmass as a
  function of zenith distance.

  @note This function can be recompiled to use the formula of Young and Irvine
        (Young A. T., Irvine W. M., 1967, Astron. J. 72, 945) instead of the
        less widely used Hardie (1962). Then the range of trustworthy airmass
        outputs is only between 1 and 4.
  @note This function can be recompiled to use the formula of Young
        (Young A. T., 1994 ApOpt, 33, 1108).
  @note http://en.wikipedia.org/w/index.php?title=Airmass&oldid=358226579#Interpolative_formulas
        has an interesting collection of other approximations.

  @error{set CPL_ERROR_ILLEGAL_INPUT\, return -1,
         aRA\, aDEC\, aLST\, or aLatitude values are not in a sensible range}
  @error{set CPL_ERROR_ILLEGAL_INPUT\, return -1, aExptime is negative}
  @error{output warning\, set CPL_ERROR_ILLEGAL_OUTPUT\, return -1,
         the computed zenith distance is invalid}
 */
/*----------------------------------------------------------------------------*/
double
muse_astro_compute_airmass(double aRA, double aDEC, double aLST,
                           double aExptime, double aLatitude)
{
  cpl_ensure(aRA >= 0. && aRA < 360. && aDEC >= -90. && aDEC <= 90. &&
             aLST >= 0. && aLST < 86400. && aLatitude >= -90. && aLatitude <= 90.,
             CPL_ERROR_ILLEGAL_INPUT, -1.);
  cpl_ensure(aExptime >= 0., CPL_ERROR_ILLEGAL_INPUT, -1.);

  /* Compute hour angle of the observation in degrees. */
  double HA = aLST * 15./3600. - aRA;
  /* Range adjustments. Angle between line of sight and the meridian is needed. */
  if (HA < -180.) {
    HA += 360.;
  }
  if (HA > 180.) {
    HA -= 360.;
  }

  /* Convert angles from degrees to radians. */
  double delta = aDEC * CPL_MATH_RAD_DEG,
         latitude = aLatitude * CPL_MATH_RAD_DEG,
         hourangle = HA * CPL_MATH_RAD_DEG;

  /* Calculate airmass of the observation using the approximation given    *
   * by Young (1994). For non-zero exposure times these airmass values are *
   * averaged using the weights given by Stetson.                          */
  double cosz = muse_astro_get_zenith_distance(hourangle, delta, latitude);
#if MUSE_ASTRO_AIRMASS_APPROX == 2
  double z = acos(cosz) * CPL_MATH_DEG_RAD;
  const double zlimit = 80.;
  if (z > zlimit) {
    cpl_msg_warning(__func__, "Zenith angle %f > %f!", z, zlimit);
  }
#endif
  if (cosz == 0. || fabs(1. / cosz) < FLT_EPSILON ||
      acos(cosz) > CPL_MATH_PI_2) {
    cpl_msg_error(__func__, "Airmass computation unsuccessful. Object is below "
                  "the horizon at start (z = %f).", acos(cosz) * CPL_MATH_DEG_RAD);
    cpl_error_set(__func__, CPL_ERROR_ILLEGAL_OUTPUT);
    return -1.;
  }

  double airmass = 0.;
#if MUSE_ASTRO_AIRMASS_APPROX == 0
  airmass = muse_astro_get_airmass_youngirvine(1. / cosz);
#endif
#if MUSE_ASTRO_AIRMASS_APPROX == 1
  airmass = muse_astro_get_airmass_young(cosz);
#endif
#if MUSE_ASTRO_AIRMASS_APPROX == 2
  airmass = muse_astro_get_airmass_hardie(1. / cosz);
#endif
#if MUSE_ASTRO_AIRMASS_APPROX > 2
#error set MUSE_ASTRO_AIRMASS_APPROX to 0, 1, 2
#endif

  /* if the exposure time is larger than zero, weight in airmass   *
   * at mid and end exposure according to Stetson's weights (which *
   * are also used in IRAF/astcalc for the eairmass function       */
  if (aExptime > 0.) {
    const double weights[] = {1./6., 2./3., 1./6.};
    const int nweights = sizeof(weights) / sizeof(double);

    double timeStep = aExptime / (nweights - 1) * 15./3600. * CPL_MATH_RAD_DEG;
    airmass *= weights[0];

    int i;
    for (i = 1; i < nweights; i++) {
      cosz = muse_astro_get_zenith_distance(hourangle + i * timeStep, delta, latitude);
#if MUSE_ASTRO_AIRMASS_APPROX == 2
      z = acos(cosz) * CPL_MATH_DEG_RAD;
      if (z > zlimit) {
        cpl_msg_warning(__func__, "Zenith angle %f > %f!", z, zlimit);
      }
#endif
      if (cosz == 0. || fabs(1. / cosz) < FLT_EPSILON ||
          acos(cosz) > CPL_MATH_PI_2) {
        cpl_msg_error(__func__, "Airmass computation unsuccessful at timeStep. "
                      "Object is below the horizon at %s exposure (z=%f).",
                      i == 1 ? "mid" : "end", acos(cosz) * CPL_MATH_DEG_RAD);
        cpl_error_set(__func__, CPL_ERROR_ILLEGAL_OUTPUT);
        return -1.;
      }

#if MUSE_ASTRO_AIRMASS_APPROX == 0
      airmass += weights[i] * muse_astro_get_airmass_youngirvine(1. / cosz);
#endif
#if MUSE_ASTRO_AIRMASS_APPROX == 1
      airmass += weights[i] * muse_astro_get_airmass_young(cosz);
#endif
#if MUSE_ASTRO_AIRMASS_APPROX == 2
      airmass += weights[i] * muse_astro_get_airmass_hardie(1. / cosz);
#endif
    } /* for i (weights) */
  } /* if aExptime */

#if MUSE_ASTRO_AIRMASS_APPROX == 0
  /* Accuracy limit for airmass approximation of Young & Irvine */
  const double airmasslimit = 4.;
  if (airmass > airmasslimit) {
    cpl_msg_warning(__func__, "Airmass larger than %f", airmasslimit);
  }
#endif

  return airmass;
} /* muse_astro_compute_airmass() */

/*----------------------------------------------------------------------------*/
/**
  @brief   Derive the effective airmass of an observation from information in
           a FITS header.
  @param   aHeader   the FITS header to use
  @return  The computed average airmass or -1 on error.

  The function queries the necessary header keywords from the input propertylist
  and calls muse_astro_compute_airmass() to do the actual computation.
  It compares the result with the ESO.TEL.AIRM.START and ESO.TEL.AIRM.END
  keywords if they are found in the input header and gives a warning if the
  output is not in between those values (with a fuzzyness of 0.005).

  @error{set CPL_ERROR_ILLEGAL_INPUT\, return -1, aHeader is NULL}
 */
/*----------------------------------------------------------------------------*/
double
muse_astro_airmass(cpl_propertylist *aHeader)
{
  cpl_ensure(aHeader, CPL_ERROR_NULL_INPUT, -1.);

  cpl_errorstate prestate = cpl_errorstate_get();
  double airm1 = muse_pfits_get_airmass_start(aHeader),
         airm2 = muse_pfits_get_airmass_end(aHeader);
  cpl_errorstate_set(prestate); /* we don't really care if these are missing */

  /* Try to read all necessary FITS headers.  Reset them to something invalid *
   * if they cannot be read, so that we get a real error message from         *
   * muse_astro_compute_airmass(), just for GEOLAT we don't care, because     *
   * muse_pfits knows a good default.                                         */
  prestate = cpl_errorstate_get();
  double ra = muse_pfits_get_ra(aHeader);
  if (!cpl_errorstate_is_equal(prestate)) {
    ra = -1000.;
  }
  prestate = cpl_errorstate_get();
  double dec = muse_pfits_get_dec(aHeader);
  if (!cpl_errorstate_is_equal(prestate)) {
    dec = -1000.;
  }
  prestate = cpl_errorstate_get();
  double lst = muse_pfits_get_lst(aHeader);
  if (!cpl_errorstate_is_equal(prestate)) {
    lst = -1000.;
  }
  prestate = cpl_errorstate_get();
  double exptime = muse_pfits_get_exptime(aHeader);
  if (!cpl_errorstate_is_equal(prestate)) {
    exptime = -1.;
  }

  double airmass = muse_astro_compute_airmass(ra, dec, lst, exptime,
                                              muse_pfits_get_geolat(aHeader));

  /* airmass computation failed, use simple average */
  if (airmass < 0. && airm1 != 0. && airm2 != 0.) {
    /* comparison doesn't make sense in this case, warn and return immediately */
    double average = (airm1 + airm2) / 2.;
    cpl_msg_warning(__func__, "airmass computation unsuccessful (%s), using simple "
                    "average of start and end values (%f)", cpl_error_get_message(),
                    average);
    return average;
  }

  /* compare computed and header values */
  cpl_msg_debug(__func__, "airmass=%f (header %f, %f)", airmass, airm1, airm2);
  if (airm1 != 0. && airm2 != 0.) {
    cpl_boolean check = airmass > fmin(airm1, airm2) - 0.005
                      && airmass < fmax(airm1, airm2) + 0.005;
    if (!check) {
      cpl_msg_warning(__func__, "Computed airmass %.3f is NOT in the range "
                      "recorded in the FITS header (%f, %f)", airmass,
                      airm1, airm2);
    }
  }

  return airmass;
} /* muse_astro_airmass() */

/*----------------------------------------------------------------------------*/
/**
  @brief   Derive the position angle of an observation from information in
           a FITS header.
  @param   aHeader   the FITS header to use
  @return  The computed position angle or 0 on error.

  The function queries the header keyword INS.DROT.POSANG and INS.DROT.MODE.
  If INS.DROT.MODE is found to be "SKY", the sign found in POSANG is inverted,
  for "STAT" it is not. For all other cases, it's not clear what should happen,
  so output a warning and leave the original sign untouched.

  @error{set CPL_ERROR_NULL_INPUT\, return 0., aHeader is NULL}
  @error{propagate error from muse_pfits_get_drot_posang(),
         INS.DROT.POSANG could not be found in the input aHeader}
  @error{propagate error from muse_pfits_get_drot_mode(),
         INS.DROT.MODE could not be found in the input aHeader}
 */
/*----------------------------------------------------------------------------*/
double
muse_astro_posangle(const cpl_propertylist *aHeader)
{
  cpl_ensure(aHeader, CPL_ERROR_NULL_INPUT, 0.);
  double posang = muse_pfits_get_drot_posang(aHeader);
  const char *mode = muse_pfits_get_drot_mode(aHeader);
  if (mode && !strncmp(mode, "SKY", 4)) {
    posang *= -1.;
  } else if (mode && strncmp(mode, "STAT", 5)) { /* not STAT, which we ignore */
    cpl_msg_warning(__func__, "Derotator mode is neither SKY nor STAT! "
                    "Effective position angle may be wrong!");
  } else if (!mode) { /* MODE missing */
    cpl_msg_warning(__func__, "Derotator mode is not given! Effective position "
                    "angle may be wrong!");
  }
  return posang;
} /* muse_astro_posangle() */

/*----------------------------------------------------------------------------*/
/**
  @brief   Properly average parallactic angle values of one exposure.
  @param   aHeader   the FITS header to use
  @return  The average parallactic angle or 0 on error.

  The function queries the header keywords ESO.TEL.PARANG.START and
  ESO.TEL.PARANG.END, and if a large flip occurs between them (i.e. from ~-180
  to ~+180) it properly averages them to a value with an absolute of ~180 not
  around zero.

  @error{set CPL_ERROR_NULL_INPUT\, return 0., aHeader is NULL}
  @error{output warning\, propagate error from muse_pfits_get_parang_start(),
         ESO.TEL.PARANG.START could not be found in the input aHeader}
  @error{output warning\, propagate error from muse_pfits_get_parang_end(),
         ESO.TEL.PARANG.END could not be found in the input aHeader}
 */
/*----------------------------------------------------------------------------*/
double
muse_astro_parangle(const cpl_propertylist *aHeader)
{
  cpl_ensure(aHeader, CPL_ERROR_NULL_INPUT, 0.);
  cpl_errorstate es = cpl_errorstate_get();
  double p1 = muse_pfits_get_parang_start(aHeader),
         p2 = muse_pfits_get_parang_end(aHeader);
  if (!cpl_errorstate_is_equal(es)) {
    cpl_msg_warning(__func__, "One or both TEL.PARANG keywords are missing!");
  }

  double parang = (p1 + p2) / 2.;
  if (fabs(p1 - p2) < 90.) {
    /* are not going through the meridian with a 360 deg flip in   *
     * PARANG, so it should be safe to just average the two values */
    return parang;
  }
  /* Flip in PARANG while observing this object, special (pretty          *
   * convoluted) handling needed: Both absolute values should be close to *
   * 180., so compute the absolute distance from 180, the average of the  *
   * distance to 180 with the respective sign, and subtract the average   *
   * of those 180. The final sign is taken from the value whose distance  *
   * to the from respectively signed 180 is larger.                       */
  double d1 = copysign(180. - fabs(p1), p1),
         d2 = copysign(180. - fabs(p2), p2);
  parang = 180. - fabs((d1 + d2) / 2.);
  if (fabs(d1) > fabs(d2)) {
    parang = copysign(parang, p1);
  } else {
    parang = copysign(parang, p2);
  } /* else */
  return parang;
} /* muse_astro_parangle() */

/*----------------------------------------------------------------------------*/
/**
  @brief   Compute angular distance in the sky between two positions.
  @param   aRA1    RA [deg] of first position
  @param   aDEC1   DEC [deg] of first position
  @param   aRA2    RA [deg] of second position
  @param   aDEC2   DEC [deg] of second position
  @return  The computed angular distance in degrees.

  This function just implements the Vincenty formula from
  http://en.wikipedia.org/w/index.php?title=Great-circle_distance&oldid=599795109#Computational_formulas
 */
/*----------------------------------------------------------------------------*/
double
muse_astro_angular_distance(double aRA1, double aDEC1, double aRA2, double aDEC2)
{
  /* first convert all inputs to radians */
  double ra1 = aRA1 * CPL_MATH_RAD_DEG,
         dec1 = aDEC1 * CPL_MATH_RAD_DEG,
         ra2 = aRA2 * CPL_MATH_RAD_DEG,
         dec2 = aDEC2 * CPL_MATH_RAD_DEG;
  /* compute formula components */
  double dra = ra2 - ra1,
         cosdra = cos(dra),
         sindec1 = sin(dec1),
         cosdec1 = cos(dec1),
         sindec2 = sin(dec2),
         cosdec2 = cos(dec2);
  /* compute the two terms in the nominator */
  double t1 = cosdec2 * sin(dra),
         t2 = cosdec1 * sindec2 - sindec1 * cosdec2 * cosdra,
         dsigma = atan2(sqrt(t1*t1 + t2*t2),
                        sindec1 * sindec2 + cosdec1 * cosdec2 * cosdra);
  return dsigma * CPL_MATH_DEG_RAD;
} /* muse_astro_angular_distance() */

/**@}*/
