/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* vim:set sw=2 sts=2 et cin: */
/*
 *
 * This file is part of the MUSE Instrument Pipeline
 * Copyright (C) 2005-2011 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*----------------------------------------------------------------------------*
 *                              Includes                                      *
 *----------------------------------------------------------------------------*/
#include <stdio.h>
#include <float.h>
#include <math.h>
#include <string.h>
#include <cpl.h>

#include "muse_processing.h"
#include "muse_instrument.h"

#include "muse_dfs.h"
#include "muse_pfits.h"
#include "muse_resampling.h"
#include "muse_utils.h"

/*----------------------------------------------------------------------------*
 *                             Debugging Macros                               *
 *         Set these to 1 or higher for (lots of) debugging output            *
 *----------------------------------------------------------------------------*/
#define DEBUG_EXPOSURES_SORT 0 /* debug muse_processing_sort_exposures() */

/*----------------------------------------------------------------------------*/
/**
 * @defgroup muse_processing   Processing framework
 *
 * The following functions are used to make the data processing in MUSE
 * easier and uniform.
 */
/*----------------------------------------------------------------------------*/

/**@{*/

/*----------------------------------------------------------------------------*
 *                              Functions                                     *
 *----------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/**
  @private
  @brief Get the raw input tag among a number of input frames.
  @param aRecipe Recipe
  @return the raw input tag, or NULL on error.

  @remark
  The information about the input tag is retrieved from the recipeconfigs.
 */
/*---------------------------------------------------------------------------*/
static const char *
muse_processing_get_rawtag(cpl_recipe *aRecipe)
{
  cpl_recipeconfig *recipeconfig = muse_processing_get_recipeconfig(aRecipe);
  if (recipeconfig == NULL) {
    cpl_msg_error(__func__, "No recipeconfig found!");
    return MUSE_TAG_EMPTY;
  }
  const char *rawTag = NULL;
  cpl_size iframe, nframes = cpl_frameset_get_size(aRecipe->frames);
  for (iframe = 0; iframe < nframes; iframe++) {
    cpl_frame *frame = cpl_frameset_get_position(aRecipe->frames, iframe);
    rawTag = cpl_frame_get_tag(frame);
    cpl_errorstate state = cpl_errorstate_get();
    char**calibTags = cpl_recipeconfig_get_inputs(recipeconfig, rawTag);
    if (calibTags != NULL) {
      int i;
      for (i = 0; calibTags[i] != NULL; i++) {
        cpl_free(calibTags[i]);
      }
      cpl_free(calibTags);
      break;
    } else {
      cpl_errorstate_set(state);
    }
  } /* for iframe (all recipe frames) */
  if (iframe >= nframes) {
    cpl_msg_error(__func__, "No valid raw tag found!");
    return MUSE_TAG_EMPTY;
  } else {
    return rawTag;
  }
}

/*---------------------------------------------------------------------------*/
/**
  @brief Create a new processing structure.
  @param aRecipeName   Name of the recipe
  @param aRecipe       the CPL recipe
  @return the processing structure
 */
/*---------------------------------------------------------------------------*/
muse_processing *
muse_processing_new(const char *aRecipeName, cpl_recipe *aRecipe)
{
  muse_processing *processing = cpl_malloc(sizeof(muse_processing));
  processing->recipeName = aRecipeName;
  processing->inputFrames = cpl_frameset_duplicate(aRecipe->frames);
  processing->recipe = aRecipe;
  processing->usedFrames = cpl_frameset_new();
  processing->outputFrames = cpl_frameset_new();
  processing->parameters = aRecipe->parameters;
  processing->inputTag = muse_processing_get_rawtag(aRecipe);
  processing->framecounter = cpl_malloc(sizeof(muse_framecounter_s));
  processing->framecounter[0].tag = NULL;
  return processing;
}

/*---------------------------------------------------------------------------*/
/**
  @brief Free the muse_processing structure.
  @param aProcessing the structure to be freed
 */
/*---------------------------------------------------------------------------*/
void
muse_processing_delete(muse_processing *aProcessing)
{
  if (!aProcessing) {
    return;
  }
  cpl_frameset_delete(aProcessing->inputFrames);
  cpl_frameset_delete(aProcessing->usedFrames);
  cpl_frameset_delete(aProcessing->outputFrames);
  int i;
  for (i = 0; aProcessing->framecounter[i].tag != NULL; i++) {
    cpl_free((char *)aProcessing->framecounter[i].tag);
  }
  cpl_free(aProcessing->framecounter);
  cpl_free(aProcessing);
}

/*---------------------------------------------------------------------------*/
/**
  @brief Add a frame to the set of used frames.
  @param aProcessing   the processing structure
  @param aFrame        the new used frame
  @param aGroup        the frame group to assign to the frame
  @param aDuplicate    if not zero, duplicate the frame before appending it
                       into the used frameset

  Insert aFrame into the usedFrames frameset of aProcessing if it is not already
  present in that frameset (present being checked using the filename and tag
  information of existing frames against aFrame). If aDuplicate is zero, the
  frame is destroyed here, assuming that it was duplicated from some other frame
  before being passed into this function. <b>In this case, any property (e.g.
  filename) queried from the frame can no longer be used afterwards!</b>
 */
/*---------------------------------------------------------------------------*/
void
muse_processing_append_used(muse_processing *aProcessing,
                            cpl_frame *aFrame, cpl_frame_group aGroup,
                            int aDuplicate)
{
  if (!aProcessing) {
    cpl_msg_error(__func__, "NULL processing struct!");
    return;
  }
#if 0
  if (aGroup == CPL_FRAME_GROUP_CALIB) {
    cpl_msg_debug(__func__, "using %s %s frame %s", "calibration",
                  cpl_frame_get_tag(aFrame), cpl_frame_get_filename(aFrame));
  } else if (aGroup == CPL_FRAME_GROUP_RAW) {
    cpl_msg_debug(__func__, "using %s %s frame %s", "raw",
                  cpl_frame_get_tag(aFrame), cpl_frame_get_filename(aFrame));
  }
#endif
  /* check, if this frame is already listed */
  const char *fn = cpl_frame_get_filename(aFrame),
             *tag = cpl_frame_get_tag(aFrame);
  cpl_size iframe, nframes = cpl_frameset_get_size(aProcessing->usedFrames);
  for (iframe = 0;
       (iframe < nframes) && fn && tag; /* only check with valid info */
       iframe++) {
    cpl_frame *frame = cpl_frameset_get_position(aProcessing->usedFrames, iframe);
    const char *fn2 = cpl_frame_get_filename(frame),
               *tag2 = cpl_frame_get_tag(frame);
    if (fn2 && !strncmp(fn, fn2, strlen(fn) + 1) &&
        tag2 && !strncmp(tag, tag2, strlen(tag) + 1)) {
      if (!aDuplicate) { /* destroy frame, assuming it was duplicated elsewhere */
        cpl_frame_delete(aFrame);
      }
      return;
    }
  } /* for iframe (all usedFrames) */
  cpl_frame_set_group(aFrame, aGroup);
  if (aDuplicate) {
    cpl_frameset_insert(aProcessing->usedFrames, cpl_frame_duplicate(aFrame));
  } else {
    cpl_frameset_insert(aProcessing->usedFrames, aFrame);
  }
}

/*---------------------------------------------------------------------------*/
/**
  @private
  @brief Get the next frame counter.
  @param aProcessing   the processing structure
  @param aTag          the frame tag
  @param aIFU          the IFU/channel number
  @return next frame counter

  Get the next frame counter. This number will start at 1 for any
  pair of tag/ifu and will increase with each call of the function
  for that pair.
 */
/*---------------------------------------------------------------------------*/
static int
muse_processing_get_framecounter(muse_processing *aProcessing,
                                 const char *aTag, int aIFU)
{
  if (!aProcessing) {
    cpl_msg_error(__func__, "NULL processing struct!");
    return 0;
  }
  int i;
  for (i = 0; aProcessing->framecounter[i].tag != NULL; i++) {
    if (strcmp(aProcessing->framecounter[i].tag, aTag) == 0 &&
        aProcessing->framecounter[i].ifu == aIFU) {
      return ++aProcessing->framecounter[i].counter;
    }
  }
  aProcessing->framecounter = cpl_realloc(aProcessing->framecounter, 
                                          (i+2)*sizeof(muse_framecounter_s));
  aProcessing->framecounter[i].tag = cpl_strdup(aTag);
  aProcessing->framecounter[i].ifu = aIFU;
  aProcessing->framecounter[i].counter = 1;
  aProcessing->framecounter[i+1].tag = NULL;
  return 1;
}

/*---------------------------------------------------------------------------*/
/**
  @private
  @brief Create a new FITS header for saving a frame.
  @param aProcessing   the processing structure
  @param aHeader       the header to be modified
  @param aFrame        the frame to be saved
  @param aIndex        input frame index (>= 0) or a negative number for master
                       or date-selected files
  @param aDateObs      the string of observed date (FITS keyword DATE-OBS), only
                       used if aIndex is negative
  @param aSequence     if target frame is part of a sequence
 */
/*---------------------------------------------------------------------------*/
static void
muse_processing_setup_header(muse_processing *aProcessing,
                             cpl_propertylist *aHeader, cpl_frame *aFrame,
                             int aIndex, const char *aDateObs,
                             cpl_boolean aSequence)
{
  if (!aProcessing) {
    cpl_msg_error(__func__, "NULL processing struct!");
    return;
  }
  if (!aProcessing->inputFrames ||
      cpl_frameset_is_empty(aProcessing->inputFrames)) {
    cpl_msg_warning(__func__, "No raw input files, no DFS product header added");
    return;
  }
  /* save some headers which are be over overwritten by cpl_dfs, *
   * including OBJECT and ESO.DRS.MUSE                           */
  cpl_propertylist *hkeep = cpl_propertylist_new();
  cpl_propertylist_copy_property_regexp(hkeep, aHeader,
                                        MUSE_HEADERS_KEEP_REGEXP, 0);
  /* don't want any DFS-generated ESO.PRO keywords before calling cpl_dfs *
   * to create them, so delete all ESO.PRO stuff and COMMENTS             */
  cpl_propertylist_erase_regexp(aHeader, "^ESO PRO|^COMMENT", 0);
  /* remove temporary stuff that is not meant to be saved to disk */
  cpl_propertylist_erase_regexp(aHeader, MUSE_HDR_TMP_REGEXP, 0);

  /* make sure that the raw frame(s) are first in the list of used frames */
  cpl_frameset *fset = muse_frameset_sort_raw_other(aProcessing->usedFrames,
                                                    aIndex, aDateObs, aSequence);

  /* XXX workaround for the EQUINOX warning, replace with double-type keyword */
  if (cpl_propertylist_has(aHeader, "EQUINOX") &&
      cpl_propertylist_get_type(aHeader, "EQUINOX") < CPL_TYPE_FLOAT) {
    cpl_property *equinox =  cpl_propertylist_get_property(aHeader, "EQUINOX");
    double equvalue = cpl_property_get_long_long(equinox);
    const char *equcomment = cpl_property_get_comment(equinox);
    cpl_property_set_name(equinox, "EQUIBRK");
    cpl_propertylist_insert_after_double(aHeader, "EQUIBRK", "EQUINOX",
                                         equvalue);
    cpl_propertylist_set_comment(aHeader, "EQUINOX", equcomment);
    cpl_propertylist_erase(aHeader, "EQUIBRK");
  } /* if non-float EQUINOX */

  char *pkgstring = cpl_sprintf("%s/%s", PACKAGE, PACKAGE_VERSION);
  if (cpl_dfs_setup_product_header(aHeader, aFrame, fset,
                                   aProcessing->parameters,
                                   aProcessing->recipeName,
                                   pkgstring, MUSE_PRO_DID, NULL)
      != CPL_ERROR_NONE) {
    cpl_msg_error(__func__, "Could not add DFS product header: %s",
                  cpl_error_get_message());
  }
  cpl_free(pkgstring);
  cpl_frameset_delete(fset);

  /* reset overwritten headers (if they were there before) */
  int i, n = cpl_propertylist_get_size(hkeep);
  for (i = 0; i < n; i++) {
    const cpl_property *p = cpl_propertylist_get_const(hkeep, i);
    cpl_propertylist_erase(aHeader, cpl_property_get_name(p));
    cpl_propertylist_append_property(aHeader, p);
  } /* for i (all properties in hkeep) */
  cpl_propertylist_delete(hkeep);

  /* MUSE data is always taken with an IFU, so override PRO.TECH */
  cpl_propertylist_update_string(aHeader, "ESO PRO TECH", "IFU");
  /* add PRO.SCIENCE header for science recipes */
  if (strstr(aProcessing->recipeName, "sci")) {
    cpl_propertylist_update_bool(aHeader, "ESO PRO SCIENCE", 1);
  }
} /* muse_processing_setup_header() */

/*---------------------------------------------------------------------------*/
/**
  @brief Create a new frame for a result file.
  @param aProcessing the processing structure
  @param aIFU        the IFU/channel number
                     (use negative value to suppress -nn postfix)
  @param aHeader     the custom properties to be included in the header
  @param aTag        tag to apply for output frames, used as default prefix
  @param aType       the type of data to be stored
  @return the new frame

  Create a new frame for a result file. Set the correct type, group, level, tag.
  The filename consists of the prefix, possibly an underscore and a running
  number (for non-unique files), a minus and the IFU number, and the suffix
  ".fits".

  As prefix, the frame tag is used on default. One may create a different
  prefix by providing the "MUSE PRIVATE FILE PREFIX" property in the header.

  All "MUSE PRIVATE " properties are removed from the header.

  @error{set CPL_ERROR_NULL_INPUT\, return NULL, aProcessing is NULL}
 */
/*---------------------------------------------------------------------------*/
cpl_frame *
muse_processing_new_frame(muse_processing *aProcessing, int aIFU,
                          cpl_propertylist *aHeader, const char *aTag,
                          cpl_frame_type aType)
{
  cpl_ensure(aProcessing, CPL_ERROR_NULL_INPUT, NULL);
  muse_processing_prepare_header(aProcessing->recipe, aTag, aHeader);
  const char *prefix = cpl_propertylist_has(aHeader, "MUSE PRIVATE FILE PREFIX")
                     ? cpl_propertylist_get_string(aHeader,
                                                   "MUSE PRIVATE FILE PREFIX")
                     : aTag;

  cpl_frame *frame = cpl_frame_new();
  /* some error from before might disturbe further error handling */
  cpl_errorstate prestate = cpl_errorstate_get();

  /* see if we need to take all input frames, a specific frame based on its  *
   * number in the frameset, and/or a specific frame based on observing date */
  int fmode = muse_processing_get_frame_mode(aProcessing->recipe, aTag),
      framecounter = 0;
  if (fmode != MUSE_FRAME_MODE_MASTER) {
    framecounter = muse_processing_get_framecounter(aProcessing, prefix, aIFU);
  }
  const char *dateobs = NULL;
  if (fmode == MUSE_FRAME_MODE_DATEOBS) {
    dateobs = muse_pfits_get_dateobs(aHeader);
  }
  cpl_boolean sequence = fmode == MUSE_FRAME_MODE_SEQUENCE;

  /* Create product frame */
  char filename[FILENAME_MAX];
  if (aIFU >= 0) {
    if (framecounter == 0) {
      snprintf(filename, FILENAME_MAX, "%s-%02d.fits", prefix, aIFU);
    } else {
      snprintf(filename, FILENAME_MAX, "%s_%04d-%02d.fits", prefix,
               framecounter, aIFU);
    }
  } else {
    if (framecounter == 0) {
      snprintf(filename, FILENAME_MAX, "%s.fits", prefix);
    } else {
      snprintf(filename, FILENAME_MAX, "%s_%04d.fits", prefix, framecounter);
    }
  }
  cpl_frame_set_filename(frame, filename);
  cpl_frame_set_tag(frame, aTag);
  cpl_frame_set_type(frame, aType);
  cpl_frame_set_group(frame, CPL_FRAME_GROUP_PRODUCT);
  cpl_frame_set_level(frame,
                      muse_processing_get_frame_level(aProcessing->recipe,
                                                      aTag));
  if (cpl_errorstate_get() != prestate) {
    cpl_msg_error(__func__, "Error while initialising the product frame: %s",
                  cpl_error_get_message());
    cpl_frame_delete(frame);
    return NULL;
  }
  cpl_propertylist_erase_regexp(aHeader, "MUSE PRIVATE.*", 0);
  /* select specific frame index to use */
  int idx = dateobs || fmode == MUSE_FRAME_MODE_SUBSET ? -1 : framecounter - 1;
  muse_processing_setup_header(aProcessing, aHeader, frame, idx, dateobs,
                               sequence);
  return frame;
} /* muse_processing_new_frame() */

/*---------------------------------------------------------------------------*/
/**
  @brief Save a computed MUSE image to disk.
  @param aProcessing the processing structure
  @param aIFU        the IFU/channel number
  @param aImage      the image to be saved
  @param aTag        tag to apply for output frames, used as prefix
  @return the CPL error code

  Save an image to disk. The filename consists on the prefix, a "-"
  sign, the IFU number of the CCD and the suffix ".fits". If the
  image saving was successful, add the output frame to the list of
  output frames in the processing structure.

  @error{CPL_ERROR_NULL_INPUT, an argument was NULL}
 */
/*---------------------------------------------------------------------------*/
int
muse_processing_save_image(muse_processing *aProcessing, int aIFU,
                           muse_image *aImage, const char *aTag)
{
  cpl_ensure_code(aProcessing && aImage && aTag, CPL_ERROR_NULL_INPUT);

  cpl_frame *frame = muse_processing_new_frame(aProcessing, aIFU,
                                               aImage->header, aTag,
                                               CPL_FRAME_TYPE_IMAGE);
  cpl_msg_info(__func__, "Saving image as %s", cpl_frame_get_filename(frame));
  int r = muse_image_save(aImage, cpl_frame_get_filename(frame));
  if (r == CPL_ERROR_NONE) {
    cpl_frameset_insert(aProcessing->outputFrames, frame);
  } else {
    cpl_frame_delete(frame);
  }
  return r;
} /* muse_processing_save_image() */

/*---------------------------------------------------------------------------*/
/**
  @brief  Save a MUSE datacube to disk.
  @param  aProcessing   the processing structure
  @param  aIFU          the IFU/channel number
  @param  aCube         the cube to be saved
  @param  aTag          tag to apply for output frames, used as default prefix
  @param  aType         type of cube to save
                        (MUSE_CUBE_TYPE_FITS or MUSE_CUBE_TYPE_EURO3D)
  @return CPL_ERROR_NONE or the relevant cpl_error_code on error

  Save a datacube to disk. The filename consists on the prefix, a "-"
  sign, the IFU number of the CCD and the suffix ".fits". If saving went without
  error, add the output frame to the list of output frames in the processing
  structure.

  @error{CPL_ERROR_NULL_INPUT, an argument was NULL}
  @error{CPL_ERROR_ILLEGAL_INPUT, an unknown aType was given}
 */
/*---------------------------------------------------------------------------*/
cpl_error_code
muse_processing_save_cube(muse_processing *aProcessing, int aIFU,
                          void *aCube, const char *aTag, muse_cube_type aType)
{
  cpl_ensure_code(aProcessing && aCube && aTag, CPL_ERROR_NULL_INPUT);
  cpl_ensure_code(aType == MUSE_CUBE_TYPE_EURO3D || aType == MUSE_CUBE_TYPE_FITS,
                  CPL_ERROR_ILLEGAL_INPUT);

  /* both cube structures contain the header component as  *
   * first element, so it's safe to cast it to one of them */
  cpl_frame *frame = muse_processing_new_frame(aProcessing, aIFU,
                                               ((muse_datacube *)aCube)->header,
                                               aTag, CPL_FRAME_TYPE_IMAGE);
  cpl_msg_info(__func__, "Saving %s cube as \"%s\"",
               aType == MUSE_CUBE_TYPE_EURO3D ? "Euro3D" : "FITS",
               cpl_frame_get_filename(frame));
  cpl_error_code rc = CPL_ERROR_NONE;
  if (aType == MUSE_CUBE_TYPE_EURO3D) {
    rc = muse_euro3dcube_save((muse_euro3dcube *)aCube,
                              cpl_frame_get_filename(frame));
  } else {
    rc = muse_datacube_save((muse_datacube *)aCube,
                            cpl_frame_get_filename(frame));
  }
  if (rc == CPL_ERROR_NONE) {
    cpl_frameset_insert(aProcessing->outputFrames, frame);
  } else {
    cpl_frame_delete(frame);
  }
  return rc;
} /* muse_processing_save_cube() */

/*---------------------------------------------------------------------------*/
/**
  @brief  Save a computed table to disk.
  @param  aProcessing   the processing structure
  @param  aIFU          the IFU/channel number
  @param  aTable        the table to be saved
  @param  aHeader       the custom properties to be included in the header
                        (for aType == MUSE_TABLE_TYPE_CPL)
  @param  aTag          tag to apply for output frames, used as prefix
  @param  aType         type of table to save
                        (MUSE_TABLE_TYPE_CPL or MUSE_TABLE_TYPE_PIXTABLE)
  @return CPL_ERROR_NONE or the relevant cpl_error_code on error

  Save a table to disk. The filename consists on the prefix, a "-"
  sign, the IFU number of the CCD and the suffix ".fits". If the
  image saving was successful, add the output frame to the list of
  output frames in the processing structure.

  @error{CPL_ERROR_NULL_INPUT, a mandatory argument was NULL}
  @error{CPL_ERROR_ILLEGAL_INPUT, an unknown aType was given}
 */
/*---------------------------------------------------------------------------*/
cpl_error_code
muse_processing_save_table(muse_processing *aProcessing, int aIFU,
                           void *aTable, cpl_propertylist *aHeader,
                           const char *aTag, muse_table_type aType)
{
  cpl_ensure_code(aProcessing && aTable && aTag, CPL_ERROR_NULL_INPUT);
  cpl_ensure_code(aType == MUSE_TABLE_TYPE_CPL || aType == MUSE_TABLE_TYPE_PIXTABLE,
                  CPL_ERROR_ILLEGAL_INPUT);

  /* for plain tables, the separate header is needed */
  cpl_propertylist *header;
  if (aType == MUSE_TABLE_TYPE_CPL) {
    cpl_ensure_code(aHeader, CPL_ERROR_NULL_INPUT);
    header = aHeader;
  } else {
    header = ((muse_pixtable *)aTable)->header;
  }
  cpl_frame *frame = muse_processing_new_frame(aProcessing, aIFU, header,
                                               aTag, CPL_FRAME_TYPE_TABLE);
  cpl_msg_info(__func__, "Saving %stable as \"%s\"",
               aType == MUSE_TABLE_TYPE_PIXTABLE ? "pixel " : "",
               cpl_frame_get_filename(frame));
  cpl_error_code rc = CPL_ERROR_NONE;
  if (aType == MUSE_TABLE_TYPE_CPL) {
    rc = cpl_table_save((cpl_table *)aTable, aHeader, NULL,
                        cpl_frame_get_filename(frame), CPL_IO_CREATE);
  } else {
    rc = muse_pixtable_save((muse_pixtable *)aTable,
                            cpl_frame_get_filename(frame));
  }
  if (rc == CPL_ERROR_NONE) {
    cpl_frameset_insert(aProcessing->outputFrames, frame);
  } else {
    cpl_msg_error(__func__, "Saving %stable failed: %s",
                  aType == MUSE_TABLE_TYPE_PIXTABLE ? "pixel " : "",
                  cpl_error_get_message());
    cpl_frame_delete(frame);
  }
  return rc;
} /* muse_processing_save_table() */

/*---------------------------------------------------------------------------*/
/**
  @brief Save a computed FITS image to disk.
  @param aProcessing the processing structure
  @param aIFU        the IFU/channel number
  @param aImage      the image to be saved
  @param aHeader     the custom properties to be included in the header
  @param aTag        tag to apply for output frames, used as prefix
  @return the CPL error code

  Save an image to disk. The filename consists of the prefix, a "-"
  sign, the IFU number of the CCD, and the suffix ".fits". If the
  image saving was successful, add the output frame to the list of
  output frames in the processing structure.

  @error{CPL_ERROR_NULL_INPUT, an argument was NULL}
 */
/*---------------------------------------------------------------------------*/
int
muse_processing_save_cimage(muse_processing *aProcessing, int aIFU,
                            cpl_image *aImage, cpl_propertylist *aHeader,
                            const char *aTag)
{
  cpl_ensure_code(aProcessing && aImage && aHeader && aTag, CPL_ERROR_NULL_INPUT);

  cpl_frame *frame = muse_processing_new_frame(aProcessing, aIFU,
                                               aHeader, aTag,
                                               CPL_FRAME_TYPE_IMAGE);
  cpl_msg_info(__func__, "Saving image as %s", cpl_frame_get_filename(frame));
  int r = cpl_image_save(aImage, cpl_frame_get_filename(frame),
                         CPL_TYPE_UNSPECIFIED, aHeader, CPL_IO_CREATE);
  if (r == CPL_ERROR_NONE) {
    cpl_frameset_insert(aProcessing->outputFrames, frame);
  } else {
    cpl_msg_error(__func__, "Saving image failed: %s", cpl_error_get_message());
    cpl_frame_delete(frame);
  }
  return r;
}

/*---------------------------------------------------------------------------*/
/**
  @brief Save a FITS header to disk.
  @param aProcessing the processing structure
  @param aIFU        the IFU/channel number
  @param aHeader     the header to be saved
  @param aTag        tag to apply for output frames, used as prefix
  @return the CPL error code

  Save a bare header to disk. The filename consists of the prefix, a "-"
  sign, the IFU number of the CCD, and the suffix ".fits". If the
  header saving was successful, add the output frame to the list of
  output frames in the processing structure.

  @error{CPL_ERROR_NULL_INPUT, an argument was NULL}
 */
/*---------------------------------------------------------------------------*/
cpl_error_code
muse_processing_save_header(muse_processing *aProcessing, int aIFU,
                            cpl_propertylist *aHeader, const char *aTag)
{
  cpl_ensure_code(aProcessing && aHeader && aTag, CPL_ERROR_NULL_INPUT);

  cpl_frame *frame = muse_processing_new_frame(aProcessing, aIFU,
                                               aHeader, aTag,
                                               CPL_FRAME_TYPE_IMAGE);
  cpl_msg_info(__func__, "Saving header as %s", cpl_frame_get_filename(frame));
  cpl_error_code rc = cpl_propertylist_save(aHeader,
                                            cpl_frame_get_filename(frame),
                                            CPL_IO_CREATE);
  if (rc == CPL_ERROR_NONE) {
    cpl_frameset_insert(aProcessing->outputFrames, frame);
  } else {
    cpl_msg_error(__func__, "Saving header failed: %s", cpl_error_get_message());
    cpl_frame_delete(frame);
  }
  return rc;
}

/*---------------------------------------------------------------------------*/
/**
  @brief Save a computed MUSE mask to disk.
  @param aProcessing the processing structure
  @param aIFU        the IFU/channel number
  @param aMask       the mask to be saved
  @param aTag        tag to apply for output frames, used as prefix
  @return the CPL error code

  Save an mask to disk. The filename consists on the prefix, a "-"
  sign, the IFU number of the CCD and the suffix ".fits". If the
  mask saving was successful, add the output frame to the list of
  output frames in the processing structure.

  @error{CPL_ERROR_NULL_INPUT, an argument was NULL}
 */
/*---------------------------------------------------------------------------*/
int
muse_processing_save_mask(muse_processing *aProcessing, int aIFU,
                          muse_mask *aMask, const char *aTag)
{
  cpl_ensure_code(aProcessing && aMask && aTag, CPL_ERROR_NULL_INPUT);

  cpl_frame *frame = muse_processing_new_frame(aProcessing, aIFU,
                                               aMask->header, aTag,
                                               CPL_FRAME_TYPE_IMAGE);
  cpl_msg_info(__func__, "Saving mask as %s", cpl_frame_get_filename(frame));
  int r = muse_mask_save(aMask, cpl_frame_get_filename(frame));
  if (r == CPL_ERROR_NONE) {
    cpl_frameset_insert(aProcessing->outputFrames, frame);
  } else {
    cpl_frame_delete(frame);
  }
  return r;
} /* muse_processing_save_mask() */

/*---------------------------------------------------------------------------*/
/**
  @brief  Check the input files for completeness.
  @param  aProcessing   the processing structure
  @param  aIFU          the IFU/channel number
  @return CPL_ERROR_NONE on success or a cpl_error_code on any error.

  Check the input frames for completeness. We check the min and max
  of all input and calibration files. On error, a message will be
  printed.

  @error{return CPL_ERROR_NULL_INPUT, aProcessing is NULL}
  @error{return CPL_ERROR_ILLEGAL_INPUT, recipe config not found in aProcessing}
  @error{output error message\, return CPL_ERROR_DATA_NOT_FOUND,
         a least one input requirement is not fulfilled}
 */
/*---------------------------------------------------------------------------*/
cpl_error_code
muse_processing_check_input(muse_processing *aProcessing, unsigned char aIFU)
{
  if (!aProcessing) {
    cpl_msg_error(__func__, "NULL processing struct");
    return CPL_ERROR_NULL_INPUT;
  }
  cpl_recipeconfig *recipeconfig
    = muse_processing_get_recipeconfig(aProcessing->recipe);
  if (!recipeconfig) {
    cpl_msg_error(__func__, "No recipeconfig found!");
    return CPL_ERROR_ILLEGAL_INPUT;
  }
  int errors = 0;

  cpl_frameset *frames = muse_frameset_find(aProcessing->inputFrames,
                                            aProcessing->inputTag, aIFU,
                                            CPL_FALSE);
  int nFrames = cpl_frameset_count_tags(frames, aProcessing->inputTag);
  cpl_frameset_delete(frames);
  int minFrames = cpl_recipeconfig_get_min_count(recipeconfig,
                                                 aProcessing->inputTag,
                                                 aProcessing->inputTag);
  int maxFrames = cpl_recipeconfig_get_max_count(recipeconfig,
                                                 aProcessing->inputTag,
                                                 aProcessing->inputTag);
  if (minFrames >= 0 && nFrames < minFrames) {
    cpl_msg_error(__func__, "Required %i, found %i input frames "
                  "with tag \"%s\" with IFU %hhu",
                  minFrames, nFrames, aProcessing->inputTag, aIFU);
    errors++;
  }
  if (maxFrames >= 0 && nFrames > maxFrames) {
    cpl_msg_error(__func__, "Maximal %i, found %i input frames "
                  "with tag \"%s\" with IFU %hhu",
                  maxFrames, nFrames, aProcessing->inputTag, aIFU);
    errors++;
  }
  char **tags = cpl_recipeconfig_get_inputs(recipeconfig,
                                            aProcessing->inputTag);
  if (tags == NULL) {
    cpl_msg_error(__func__, "Input frames %s cannot be used with this recipe",
                  aProcessing->inputTag);
    errors++;
  }
  int i;
  for (i = 0; tags != NULL && tags[i] != NULL; i++) {
    frames = muse_frameset_find(aProcessing->inputFrames, tags[i], aIFU,
                                CPL_FALSE);
    nFrames = cpl_frameset_count_tags(frames, tags[i]);
    cpl_frameset_delete(frames);

    minFrames = cpl_recipeconfig_get_min_count(recipeconfig,
                                               aProcessing->inputTag,
                                               tags[i]);
    maxFrames = cpl_recipeconfig_get_max_count(recipeconfig,
                                               aProcessing->inputTag,
                                               tags[i]);
    if (minFrames >= 0 && nFrames < minFrames) {
      cpl_msg_error(__func__, "Required %i, found %i frames "
                    "with tag \"%s\" with IFU %hhu",
                    minFrames, nFrames, tags[i], aIFU);
      errors++;
    }
    if (nFrames == 0 && minFrames <= 0) {
      cpl_msg_debug(__func__, "Optional frame with tag %s not given", tags[i]);
    }
    if (maxFrames >= 0 && nFrames > maxFrames) {
      cpl_msg_error(__func__, "Maximal %i, found %i frames "
                    "with tag \"%s\" with IFU %hhu",
                    maxFrames, nFrames, tags[i], aIFU);
      errors++;
    }
    cpl_free(tags[i]);
  }
  cpl_free(tags);

  if (errors > 0) {
    cpl_msg_error(__func__, "Found %i error(s)", errors);
    return CPL_ERROR_DATA_NOT_FOUND;
  }
  return CPL_ERROR_NONE;
} /* muse_processing_check_input() */

/*----------------------------------------------------------------------------*/
/**
  @brief    Sort input frames (containing lists of pixel table filenames) into
            different exposures.
  @param    aProcessing   the processing structure, especially containing
                          input/output frames and the tag
  @return   the exposure table on success or NULL on error

  @remark   This function assumes that all relevant FITS keywords (DATE-OBS
            and EXTNAME with the CHANnn number) are contained in the primary
            header.

  This function uses the DATE-OBS header keyword to check which of the input
  objects belong to the same exposure. As they all have been split up from the
  same exposure in the beginning, DATE-OBS should be identical for related
  files.

  If one IFU of the same exposure is given multiple times in the input
  frameset, only the first one is used and a warning is printed for all others.

  @error{set CPL_ERROR_NULL_INPUT\, return NULL, invalid processing pointer}
  @error{set CPL_ERROR_DATA_NOT_FOUND\, return NULL, no exposures found}
 */
/*----------------------------------------------------------------------------*/
cpl_table *
muse_processing_sort_exposures(muse_processing *aProcessing)
{
  cpl_ensure(aProcessing, CPL_ERROR_NULL_INPUT, NULL);
  cpl_size nframes = cpl_frameset_get_size(aProcessing->inputFrames);
  cpl_ensure(nframes, CPL_ERROR_DATA_NOT_FOUND, NULL);

  /* now that we have at least one frame, we can assume that the frameset is   *
   * valid and we have at least one exposure, so we can create the empty table */
  cpl_table *exptable = cpl_table_new(0);
  /* create columns for DATE-OBS and all IFUs */
  cpl_table_new_column(exptable, "DATE-OBS", CPL_TYPE_STRING);
  char colname[3]; /* further used in this function, define outside loop */
  int i;
  for (i = 0; i <= kMuseNumIFUs; i++) {
    snprintf(colname, 3, "%02d", i);
    cpl_table_new_column(exptable, colname, CPL_TYPE_STRING);
  }

  /* loop over all input frames */
  cpl_size iframe;
  for (iframe = 0; iframe < nframes; iframe++) {
    cpl_frame *frame = cpl_frameset_get_position(aProcessing->inputFrames,
                                                 iframe);
    if (!strcmp(cpl_frame_get_tag(frame), aProcessing->inputTag)) {
      const char *filename = cpl_frame_get_filename(frame);

      /* load the primary header and get DATE-OBS and the IFU number */
      cpl_propertylist *header = cpl_propertylist_load(filename, 0);
      const char *date = muse_pfits_get_dateobs(header);
      if (!date) {
        cpl_msg_warning(__func__, "\"%s\" does not contain the DATE-OBS "
                        "keyword, it will be ignored!", filename);
        cpl_propertylist_delete(header);
        continue; /* skip on to the next frame */
      }
      int ifu = muse_utils_get_ifu(header);
      if (!ifu) {
        cpl_msg_debug(__func__, "\"%s\" seems to contain merged data (no "
                      "EXTNAME=CHANnn)", filename);
      }
#if DEBUG_EXPOSURES_SORT
      cpl_msg_debug(__func__, "\"%s\": IFU %2d, DATE-OBS=\"%s\"", filename, ifu, date);
#endif

      /* loop through all exposures (=rows) in the table and compare the date */
      int irow = -1;
      for (i = 0; i < cpl_table_get_nrow(exptable); i++) {
#if DEBUG_EXPOSURES_SORT
        cpl_msg_debug(__func__, "i=%d, DATE-OBS=\"%s\"", i,
                      cpl_table_get_string(exptable, "DATE-OBS", i));
#endif
        if (!strcmp(date, cpl_table_get_string(exptable, "DATE-OBS", i))) {
          irow = i;
        }
      }

      /* didn't find a matching entry, add a new exposure (=row) */
      if (irow < 0) {
        cpl_table_set_size(exptable, cpl_table_get_nrow(exptable)+1);
        irow = cpl_table_get_nrow(exptable) - 1;
        cpl_table_set_string(exptable, "DATE-OBS", irow, date);
      }

      /* check, if table cell for this exposure is still empty and, *
       * if so, add the filename                                    */
      snprintf(colname, 3, "%02d", ifu);
      if (cpl_table_is_valid(exptable, colname, irow)) {
        cpl_msg_warning(__func__, "we already have IFU %d of exposure %d (\"%s\")!"
                        " Ignoring \"%s\"", ifu, irow+1,
                        cpl_table_get_string(exptable, colname, irow),
                        filename);
        cpl_propertylist_delete(header);
        continue; /* skip on to the next frame */
      }
      cpl_table_set_string(exptable, colname, irow, filename);
      /* now we can assume that this is valid and will be used; it is not a   *
       * RAW file, but it's the main input, so the RAW group will satisfy the *
       * DFS-related functions in CPL                                         */
      muse_processing_append_used(aProcessing, frame, CPL_FRAME_GROUP_RAW, 1);

      cpl_propertylist_delete(header);
    } /* if correct tag */
#if DEBUG_EXPOSURES_SORT
    else {
      cpl_msg_debug(__func__, "wrong tag \"%s\" for \"%s\"",
                    cpl_frame_get_tag(frame), cpl_frame_get_filename(frame));
    }
#endif
  } /* for iframe (all input frames) */

  if (cpl_table_get_nrow(exptable) <= 0) {
    /* no exposures found */
    cpl_table_delete(exptable);
    cpl_error_set(__func__, CPL_ERROR_DATA_NOT_FOUND);
    return NULL;
  }
  /* loop through all exposures and print the number of valid IFUs found */
  for (i = 0; i < cpl_table_get_nrow(exptable); i++) {
    int nmerged = 0;
    if (cpl_table_is_valid(exptable, "00", i)) {
      nmerged++;
    }
    int j, nvalid = 0;
    for (j = 1; j <= kMuseNumIFUs; j++) {
      snprintf(colname, 3, "%02d", j);
      if (cpl_table_is_valid(exptable, colname, i)) {
        nvalid++;
      }
    } /* for i (table columns) */
    cpl_msg_debug(__func__, "Data from exposure %2d is contained in %2d "
                  "IFU%s/%d merged file%s", i+1, nvalid, nvalid == 1 ? "" : "s",
                  nmerged, nmerged == 1 ? "" : "s");
  } /* for i (table rows) */

  /* sort table by increasing DATE-OBS, so that we process the exposures *
   * in chronological order                                              */
  cpl_propertylist *sorting = cpl_propertylist_new();
  cpl_propertylist_append_bool(sorting, "DATE-OBS",
                               CPL_FALSE); /* sort ascending */
  cpl_table_sort(exptable, sorting);
  cpl_propertylist_delete(sorting);

  return exptable;
} /* muse_processing_sort_exposures() */

/*---------------------------------------------------------------------------*/
/**
  @brief  Load a mask file and its FITS header
  @param  aProcessing   the processing structure, includes the input frames list
  @param  aTag          the tag
  @return a new <tt><b>muse_mask *</b></tt> or NULL on error
  @remark The new mask has to be deallocated using
          <tt><b>muse_mask_delete()</b></tt>.
  @remark Only FITS keywords from the primary FITS header will be loaded.

  @error{return NULL\, propagate CPL error code, cpl_propertylist_load
         fails (this handles the case of an invalid or empty filename) }
 */
/*---------------------------------------------------------------------------*/
muse_mask *
muse_processing_mask_load(muse_processing *aProcessing, const char *aTag)
{
  cpl_frameset *frames = muse_frameset_find(aProcessing->inputFrames,
                                            aTag, 0, CPL_FALSE);
  if (frames == NULL || cpl_frameset_get_size(frames) <= 0) {
    cpl_frameset_delete(frames);
    return NULL;
  }
  cpl_frame *frame = cpl_frameset_get_position(frames, 0);
  muse_mask *skymask = muse_mask_load(cpl_frame_get_filename(frame));
  cpl_msg_info(__func__, "using sky mask \"%s\" (%"CPL_SIZE_FORMAT" pixels)",
               cpl_frame_get_filename(frame), cpl_mask_count(skymask->mask));
  muse_processing_append_used(aProcessing, frame, CPL_FRAME_GROUP_CALIB, 1);
  cpl_frameset_delete(frames);
  return skymask;
}

/**@}*/
