/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* vim:set sw=2 sts=2 et cin: */
/*
 *
 * This file is part of the MUSE Instrument Pipeline
 * Copyright (C) 2005-2011 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

/*----------------------------------------------------------------------------*
 *                               Includes                                     *
 *----------------------------------------------------------------------------*/
#include <muclipm_make_image.h>
#include <muse.h>
#include "muse_quick_image_z.h"

/*---------------------------------------------------------------------------*
 *                             Functions code                                *
 *---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/**
  @brief Compute the quick reconstructed image and save it to disk.
  @param    aProcessing  the processing structure
  @param    aParams      the parameters list
  @return 0 on success, -1 on error

  Compared to other recipes, muse_quick_image is special-cased in
  xmldoc/recipedoc2c.xsl so that aParams->nifu stays zero if zero was passed to
  the recipe.
 */
/*---------------------------------------------------------------------------*/
int
muse_quick_image_compute(muse_processing *aProcessing,
                         muse_quick_image_params_t *aParams)
{
  /* local variables */
  cpl_image *raw_img, *bias, *flat, *mask;
  cpl_propertylist *mask_hdr;
  cpl_image *computed_img;
  cpl_propertylist *output_header = NULL;
  int nifu, n1, n2;
  long xstart[MUCLIPM_NB_SLICES*MUCLIPM_NB_SUBSLICERS],
       xend[MUCLIPM_NB_SLICES*MUCLIPM_NB_SUBSLICERS];

  /* check the nifu parameter */
  if (aParams->nifu < 0 || aParams->nifu > kMuseNumIFUs) {
    cpl_msg_error(__func__, "Illegal IFU number %d", aParams->nifu);
    return -1;
  }

  /* create the ouput image */
  computed_img = cpl_image_new(MUCLIPM_NB_SUBSLICERS * MUCLIPM_NB_SPEC_PER_SLICE,
                               MUCLIPM_NB_SLICES * MUCLIPM_NB_IFU,
                               CPL_TYPE_FLOAT);

  n1 = aParams->nifu ? aParams->nifu : 1,
  n2 = aParams->nifu ? aParams->nifu : kMuseNumIFUs;
  for (nifu = n1; nifu <= n2; nifu++) {
    int extension;
    cpl_frameset *input_frames;
    cpl_frame *frame;
    cpl_image *ifu_img;

    cpl_msg_info(__func__, "Processing IFU %d", nifu);

    /* duplicate the frameset so that we can delete files */
    input_frames = cpl_frameset_duplicate(aProcessing->inputFrames);

    frame = cpl_frameset_find(input_frames, MUSE_TAG_MASK_IMAGE);
    extension = muse_utils_get_extension_for_ifu(cpl_frame_get_filename(frame),
                                                 nifu);
    mask = cpl_image_load(cpl_frame_get_filename(frame), CPL_TYPE_INT, 0,
                          extension);
    if (!mask) {
      if (extension < 0) { /* append to CPL error message */
        cpl_error_set_message(__func__, cpl_error_get_code(), "extension for "
                              "IFU %d not found", nifu);
      }
      cpl_msg_error(__func__, "Could not load frame %s (%s): %s",
                    cpl_frame_get_filename(frame), MUSE_TAG_MASK_IMAGE,
                    cpl_error_get_message());
      cpl_frameset_delete(input_frames);
      continue; /* try the next IFU */
    }
    /* if reading of the image succeeded, there is no reason why loading *
     * the header should fail, so no need to test for success            */
    mask_hdr = cpl_propertylist_load(cpl_frame_get_filename(frame), extension);
    muse_processing_append_used(aProcessing, frame, CPL_FRAME_GROUP_CALIB, 1);
    cpl_frameset_erase(input_frames, MUSE_TAG_MASK_IMAGE);

    /* read slice limitder */
    long i;
    for (i = 0; i < MUCLIPM_NB_SLICES*MUCLIPM_NB_SUBSLICERS; i++) {
      char *p1 = cpl_sprintf("ESO DET SLICE%ld XSTART", i + 1),
           *p2 = cpl_sprintf("ESO DET SLICE%ld XEND", i + 1);
      xstart[i] = cpl_propertylist_get_int(mask_hdr, p1);
      xend[i] = cpl_propertylist_get_int(mask_hdr, p2);
      cpl_free(p1);
      cpl_free(p2);
#if 0 /* DEBUG */
      cpl_msg_debug(__func__, "Slice %ld: limits [%ld,%ld]", i + 1,
                    xstart[i], xend[i]);
#endif
      if (xstart[i] >= xend[i]) {
        cpl_msg_error(__func__, "IFU %d: inconsistant slice limits in X "
                      "[%ld,%ld] for slice %ld", nifu,
                      xstart[i], xend[i], i + 1);
        return -1;
      } /* if */
    } /* for i (all slices) */
    for (i = 1; i < MUCLIPM_NB_SLICES*MUCLIPM_NB_SUBSLICERS; i++) {
      if (xstart[i] <= xend[i-1]) {
        cpl_msg_error(__func__, "IFU %d: inconsistant interslice for slices "
                      "%ld and %ld (%ld,%ld)", nifu,
                      i - 1, i, xend[i-1], xstart[i]);
        return -1;
      } /* if */
    } /* for i (all but the first slice) */

    cpl_errorstate prestate = cpl_errorstate_get();
    frame = cpl_frameset_find(input_frames, MUSE_TAG_BIAS);
    extension = muse_utils_get_extension_for_ifu(cpl_frame_get_filename(frame),
                                                 nifu);
    bias = cpl_image_load(cpl_frame_get_filename(frame),
                          CPL_TYPE_FLOAT, 0, extension);
    if (!bias) {
      cpl_msg_debug(__func__, "Proceeding without %s that could not be loaded: "
                    "%s", MUSE_TAG_BIAS, cpl_error_get_message());
      cpl_errorstate_set(prestate); /* clean up non-critical error */
    } else {
      muse_processing_append_used(aProcessing, frame, CPL_FRAME_GROUP_CALIB, 1);
    }
    muse_processing_append_used(aProcessing, frame, CPL_FRAME_GROUP_CALIB, 1);
    cpl_frameset_erase(input_frames, MUSE_TAG_BIAS);

    prestate = cpl_errorstate_get();
    frame = cpl_frameset_find(input_frames, MUSE_TAG_FLAT);
    extension = muse_utils_get_extension_for_ifu(cpl_frame_get_filename(frame),
                                                 nifu);
    flat = cpl_image_load(cpl_frame_get_filename(frame), CPL_TYPE_FLOAT, 0,
                          extension);
    if (!flat) {
      cpl_msg_debug(__func__, "Proceeding without %s that could not be loaded: "
                    "%s", MUSE_TAG_FLAT, cpl_error_get_message());
      cpl_errorstate_set(prestate); /* clean up non-critical error */
    } else {
      muse_processing_append_used(aProcessing, frame, CPL_FRAME_GROUP_CALIB, 1);
    }
    cpl_frameset_erase(input_frames, MUSE_TAG_FLAT);

    /* the remaining frame should be the raw data */
    frame = cpl_frameset_get_position(input_frames, 0);
    extension = muse_utils_get_extension_for_ifu(cpl_frame_get_filename(frame),
                                                 nifu);
    raw_img = cpl_image_load(cpl_frame_get_filename(frame), CPL_TYPE_FLOAT, 0,
                             extension);
    if (!raw_img) {
      if (extension < 0) { /* append to CPL error message */
        cpl_error_set_message(__func__, cpl_error_get_code(), "extension for "
                              "IFU %d not found", nifu);
      }
      cpl_msg_error(__func__, "Could not load data frame %s: %s",
                    cpl_frame_get_filename(frame), cpl_error_get_message());
      cpl_frameset_delete(input_frames);
      cpl_image_delete(mask);
      cpl_propertylist_delete(mask_hdr);
      cpl_image_delete(bias);
      cpl_image_delete(flat);
      continue; /* try the next IFU */
    }
    muse_processing_append_used(aProcessing, frame, CPL_FRAME_GROUP_RAW, 1);
    /* read output header from input file */
    if (!output_header) {
      output_header = cpl_propertylist_load(cpl_frame_get_filename(frame), 0);
    }
    cpl_frameset_delete(input_frames);

    /* compute reconstructed image for this one IFU */
    ifu_img = cpl_image_new(MUCLIPM_NB_SUBSLICERS * MUCLIPM_NB_SPEC_PER_SLICE,
                            MUCLIPM_NB_SLICES, CPL_TYPE_FLOAT);
    if (muclipm_make_image(raw_img, bias, flat, mask, xstart, xend, &ifu_img)
        != CPL_ERROR_NONE) {
      cpl_msg_info(__func__, "Error while computing reconstructed image: %s",
                   cpl_error_get_message());
      cpl_image_delete(computed_img);
      cpl_propertylist_delete(output_header);
      cpl_image_delete(mask);
      cpl_propertylist_delete(mask_hdr);
      cpl_image_delete(bias);
      cpl_image_delete(flat);
      return -1; /* this should really be fatal */
    }

    /* copy reconstructed image of this IFU into the output image */
    cpl_image_copy(computed_img, ifu_img, 1, (24 - nifu) * MUCLIPM_NB_SLICES + 1);

    /* free memory associated to images, these are NULL tolerant... */
    cpl_image_delete(bias);
    cpl_image_delete(flat);
    cpl_image_delete(mask);
    cpl_propertylist_delete(mask_hdr);
    cpl_image_delete(raw_img);
    cpl_image_delete(ifu_img);
  } /* for nifu */

  cpl_image_flip(computed_img, 0);
  muse_processing_save_cimage(aProcessing, aParams->nifu, computed_img,
                              output_header, MUSE_TAG_QUICK_IMAGE);
  cpl_image_delete(computed_img);
  cpl_propertylist_delete(output_header);

  return 0;
} /* muse_quick_image_compute() */
