/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* vim:set sw=2 sts=2 et cin: */
/*
 * This file is part of the MUSE Instrument Pipeline
 * Copyright (C) 2005-2014 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*---------------------------------------------------------------------------*
 *                             Includes                                      *
 *---------------------------------------------------------------------------*/
#include <muse.h>
#include "muse_scipost_combine_pixtables_z.h"

/*----------------------------------------------------------------------------*/
/**
  @brief    Combine several pixel tables and exposures
  @param    aProcessing  the processing structure
  @param    aParams      the parameters list
  @return   0 if everything is ok, -1 something went wrong

  @see muse_scipost_compute()
 */
/*----------------------------------------------------------------------------*/
int
muse_scipost_combine_pixtables_compute(muse_processing *aProcessing,
                                       muse_scipost_combine_pixtables_params_t *aParams)
{
  /* setup and check weighting scheme */
  int weight = MUSE_XCOMBINE_EXPTIME;
  if (aParams->weight == MUSE_SCIPOST_COMBINE_PIXTABLES_PARAM_WEIGHT_FWHM) {
    weight = MUSE_XCOMBINE_FWHM;
  } else if (aParams->weight != MUSE_SCIPOST_COMBINE_PIXTABLES_PARAM_WEIGHT_EXPTIME) {
    /* it's _not_ the other one, either... */
    cpl_msg_error(__func__, "unknown weighting method \"%s\"",
                  aParams->weight_s);
    return -1;
  }

  /* sort input pixel tables into different exposures */
  cpl_table *exposures = muse_processing_sort_exposures(aProcessing);
  if (!exposures) {
    cpl_msg_error(__func__, "no science exposures found in input");
    return -1;
  }
  int nexposures = cpl_table_get_nrow(exposures);

  /* now process all the pixel tables, do it separately for each exposure */
  /* allocate one additional element for NULL termination */
  muse_pixtable **pixtables = cpl_calloc(nexposures + 1, sizeof(muse_pixtable *));
  int i;
  for (i = 0; i < nexposures; i++) {
    cpl_table *thisexp = cpl_table_extract(exposures, i, 1);
    pixtables[i] = muse_pixtable_load_merge_channels(thisexp,
                                                     aParams->lambdamin,
                                                     aParams->lambdamax);
    cpl_table_delete(thisexp);
    /* erase pre-existing QC parameters */
    cpl_propertylist_erase_regexp(pixtables[i]->header, "ESO QC ", 0);
  }
  cpl_table_delete(exposures);

  do { /* loop to break out of in case of error */
    /* now combine the possibly more than one exposures */
    muse_pixtable *bigpixtable = NULL;
    if (nexposures > 1) {
      int rc = muse_xcombine_weights(pixtables, weight);
      if (rc != CPL_ERROR_NONE) {
        cpl_msg_error(__func__, "weighting the pixel tables didn't work: %s",
                      cpl_error_get_message());
        break;
      }
      /* combine individual pixel tables and delete them */
      bigpixtable = muse_xcombine_tables(pixtables);
      if (!bigpixtable) {
        cpl_msg_error(__func__, "combining the pixel tables didn't work: %s",
                      cpl_error_get_message());
        break;
      }
    } else {
      bigpixtable = pixtables[0];
    }
    cpl_free(pixtables);

    muse_processing_save_table(aProcessing, -1, bigpixtable, NULL,
                               MUSE_TAG_PIXTABLE_COMBINED,
                               MUSE_TABLE_TYPE_PIXTABLE);
    muse_pixtable_delete(bigpixtable);
    return 0;
  } while (0); /* end of "error-free" loop, error handling follows */
  for (i = 0; i < nexposures; i++) {
    muse_pixtable_delete(pixtables[i]);
  } /* for i (exposures) */
  cpl_free(pixtables);
  return -1;
} /* muse_scipost_combine_pixtables_compute() */
