/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* vim:set sw=2 sts=2 et cin: */
/*
 *
 * This file is part of the MUSE Instrument Pipeline
 * Copyright (C) 2005-2011 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*---------------------------------------------------------------------------*
 *                             Includes                                      *
 *---------------------------------------------------------------------------*/
#include <string.h>
#include <math.h>

#include <muse.h>
#include "muse_sky_lsf_z.h"

/*---------------------------------------------------------------------------*
 *                             Functions code                                *
 *---------------------------------------------------------------------------*/
static muse_pixtable *
muse_sky_lsf_load_pixtable(muse_processing *aProcessing,
                           muse_sky_lsf_params_t *aParams)
{
  cpl_frameset *frames = muse_frameset_find(aProcessing->inputFrames,
                                            "PIXTABLE_ARC", 0, CPL_FALSE);
  if (frames == NULL || cpl_frameset_get_size(frames) <= 0) {
    cpl_frameset_delete(frames);
    return NULL;
  }
  cpl_frame *frame = cpl_frameset_get_position(frames, 0);
  const char *fn = cpl_frame_get_filename(frame);
  double lambdamin = aParams->lambdamin - kMuseSpectralSamplingA,
         lambdamax = aParams->lambdamax + kMuseSpectralSamplingA;
  muse_pixtable *pixtable = muse_pixtable_load_restricted_wavelength(fn,
                                                                     lambdamin,
                                                                     lambdamax);
  cpl_msg_info(__func__, "using pixel table \"%s\"", fn);
  muse_processing_append_used(aProcessing, frame, CPL_FRAME_GROUP_RAW, 1);
  cpl_frameset_delete(frames);
  if (pixtable != NULL) {
    cpl_table_and_selected_int(pixtable->table, MUSE_PIXTABLE_DQ,
                               CPL_NOT_EQUAL_TO, EURO3D_GOODPIXEL);
    cpl_table_erase_selected(pixtable->table);
    /* erase pre-existing QC parameters */
    cpl_propertylist_erase_regexp(pixtable->header, "ESO QC ", 0);
  }
  return pixtable;
}

static cpl_table *
muse_sky_lsf_load_lines(muse_processing *aProcessing)
{
  cpl_frameset *frames = muse_frameset_find(aProcessing->inputFrames,
                                            "SKY_LINES", 0, CPL_FALSE);
  if (frames == NULL || cpl_frameset_get_size(frames) <= 0) {
    cpl_frameset_delete(frames);
    return NULL;
  }
  cpl_frame *frame = cpl_frameset_get_position(frames, 0);
  cpl_table *lines = cpl_table_load(cpl_frame_get_filename(frame), 1, 1);
  cpl_msg_info(__func__, "using line table %s", cpl_frame_get_filename(frame));
  muse_processing_append_used(aProcessing, frame, CPL_FRAME_GROUP_CALIB, 1);
  cpl_frameset_delete(frames);
  return lines;
}

static void
muse_sky_lsf_save_params(cpl_propertylist *aHeader,
                         const muse_lsf_params **aSlicePars,
                         muse_processing *aProcessing)
{
  if (aSlicePars != NULL) {
    cpl_frame *frame 
      = muse_processing_new_frame(aProcessing, -1, aHeader,
                                  "LSF_TABLE", CPL_FRAME_TYPE_TABLE);
    if (frame != NULL) {
      const char *filename = cpl_frame_get_filename(frame);
      int r;
      r = cpl_propertylist_save(aHeader, filename, CPL_IO_CREATE);
      r = muse_lsf_params_save(aSlicePars, filename);
      if (r == CPL_ERROR_NONE) {
        cpl_frameset_insert(aProcessing->outputFrames, frame);
      } else {
        cpl_frame_delete(frame);
      }
    }
  }

}

static muse_lsf_params *
muse_sky_lsf_compute_slice_params(cpl_table *aLines,
                                  muse_pixtable *aPixtable,
                                  muse_lsf_params *aFirstguess,
                                  int max_iterations,
                                  const muse_sky_fit_params *slice_fit_params)

{
  uint32_t origin = (uint32_t)cpl_table_get_int(aPixtable->table,
                                                MUSE_PIXTABLE_ORIGIN, 0, NULL);
  int i_ifu = muse_pixtable_origin_get_ifu(origin);
  int i_slice = muse_pixtable_origin_get_slice(origin);
  cpl_size nrows = muse_pixtable_get_nrow(aPixtable);

  cpl_msg_info(__func__, "processing slice %i.%i with %"CPL_SIZE_FORMAT" rows",
               i_ifu, i_slice, nrows);

  cpl_propertylist *order = cpl_propertylist_new();
  cpl_propertylist_append_bool(order, MUSE_PIXTABLE_LAMBDA, CPL_FALSE);
  cpl_table_sort(aPixtable->table, order);
  cpl_propertylist_delete(order);

  cpl_array *lambda = NULL;
  if (cpl_table_get_column_type(aPixtable->table, MUSE_PIXTABLE_LAMBDA)
      == CPL_TYPE_DOUBLE) {
    lambda = muse_cpltable_extract_column(aPixtable->table,
                                          MUSE_PIXTABLE_LAMBDA);
  } else {
    cpl_table_cast_column(aPixtable->table, MUSE_PIXTABLE_LAMBDA,
                          "lambda_double", CPL_TYPE_DOUBLE);
    lambda = muse_cpltable_extract_column(aPixtable->table, "lambda_double");
  }
  cpl_array *data = muse_cpltable_extract_column(aPixtable->table,
                                                 MUSE_PIXTABLE_DATA);
  cpl_array *stat = muse_cpltable_extract_column(aPixtable->table,
                                                 MUSE_PIXTABLE_STAT);
  
  muse_lsf_params *lsfParam 
    = muse_lsf_params_fit(lambda, data, stat, aLines, aFirstguess, 
                                   max_iterations, slice_fit_params);
  lsfParam->slice = i_slice;
  lsfParam->ifu = i_ifu;

  if (cpl_table_has_column(aPixtable->table, "lambda_double")) {
    cpl_table_erase_column(aPixtable->table, "lambda_double");
  }

  cpl_array_unwrap(lambda);
  cpl_array_unwrap(data);
  cpl_array_unwrap(stat);

  return lsfParam;
}

static muse_lsf_params **
muse_sky_lsf_compute_slices(cpl_table *aLines,
                            muse_pixtable *aPixtable,
                            muse_lsf_params **aFirstguess,
                            int max_iterations,
                            const muse_sky_fit_params *slice_fit_params)
{
  muse_pixtable **slice_pixtable = muse_pixtable_extracted_get_slices(aPixtable);
  int n_slices = muse_pixtable_extracted_get_size(slice_pixtable);
  int i_slice;

  muse_lsf_params **lsfParams 
    = cpl_calloc(n_slices+1, sizeof(muse_lsf_params *));
#if 1
  #pragma omp parallel for default(none)               /* as req. by Ralf */ \
    shared(aLines, n_slices, i_slice, lsfParams, \
           slice_pixtable, aFirstguess, max_iterations, slice_fit_params)
#endif
  for (i_slice = 0; i_slice < n_slices; i_slice++) {
    uint32_t origin 
      = (uint32_t)cpl_table_get_int(slice_pixtable[i_slice]->table,
                                    MUSE_PIXTABLE_ORIGIN, 0, NULL);
    int ifu = muse_pixtable_origin_get_ifu(origin);
    int slice = muse_pixtable_origin_get_slice(origin);
    muse_lsf_params *firstguess_slice 
      = muse_lsf_params_get(aFirstguess, ifu, slice); 

    cpl_errorstate prestate = cpl_errorstate_get();
    lsfParams[i_slice]
      = muse_sky_lsf_compute_slice_params(aLines, slice_pixtable[i_slice],
                                          firstguess_slice, max_iterations,
                                          slice_fit_params);
    if (!cpl_errorstate_is_equal(prestate)) {
      cpl_msg_error(__func__, "While processing slice %i.%i:", ifu, slice);
      cpl_errorstate_dump(prestate, CPL_FALSE, NULL);
      cpl_errorstate_set(prestate);
    }
  }  
  
  muse_pixtable_extracted_delete(slice_pixtable);
  return lsfParams;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Interpret the command line options and execute the data processing
  @param    aProcessing   the processing structure
  @param    aParams       the parameters list
  @return   0 if everything is ok, -1 something went wrong
 */
/*----------------------------------------------------------------------------*/
int
muse_sky_lsf_compute(muse_processing *aProcessing,
                     muse_sky_lsf_params_t *aParams)
{
  cpl_msg_debug(__func__, "recipe=\"%s\", tag=\"%s\"", aProcessing->recipeName,
                aProcessing->inputTag);

  cpl_errorstate prestate = cpl_errorstate_get();

  muse_pixtable *pixtable = muse_sky_lsf_load_pixtable(aProcessing, aParams);
  if (pixtable == NULL) {
    return -1;
  }

  if (aParams->orig != NULL) {
    cpl_table_duplicate_column(pixtable->table, aParams->orig, 
                               pixtable->table, "data");
  }
  cpl_table *lines = muse_sky_lsf_load_lines(aProcessing);
  if (lines == NULL) {
    muse_pixtable_delete(pixtable);
    return -1;
  }

  muse_sky_fit_params *slice_fit_params = muse_sky_fit_params_new
    (
     aParams->slice_fit_offset,
     aParams->slice_fit_refraction,
     0, // sensitivity is not used here
     aParams->slice_fit_slit_width,
     aParams->slice_fit_bin_width,
     aParams->slice_fit_lsf_width + 1,
     aParams->slice_fit_h3 + 1, 
     aParams->slice_fit_h4 + 1, 
     aParams->slice_fit_h5 + 1, 
     aParams->slice_fit_h6 + 1
     );

  muse_lsf_params **firstguess = muse_processing_lsf_params_load(aProcessing, 0);
  muse_lsf_params **lsf = muse_sky_lsf_compute_slices(lines, pixtable, firstguess,
                                                      aParams->max_iterations,
                                                      slice_fit_params);
  muse_lsf_params_delete(firstguess);
  muse_sky_fit_params_delete(slice_fit_params);

  cpl_propertylist *qc_header = cpl_propertylist_new();
  muse_sky_lsf_save_params(qc_header, (const muse_lsf_params **)lsf, aProcessing);
  muse_lsf_params_delete(lsf);

  if (aParams->store_subtracted) {
    muse_processing_save_table(aProcessing, -1, pixtable, NULL,
                               MUSE_TAG_PIXTABLE_REDUCED,
                               MUSE_TABLE_TYPE_PIXTABLE);
  }
  /*
    Clean up the local objects.
   */
  cpl_propertylist_delete(qc_header);
  muse_pixtable_delete(pixtable);

  return (cpl_errorstate_is_equal(prestate))?0:-1;
}
