/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* vim:set sw=2 sts=2 et cin: */
/*
 *
 * This file is part of the MUSE Instrument Pipeline
 * Copyright (C) 2005-2011 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*----------------------------------------------------------------------------*
 *                              Includes                                      *
 *----------------------------------------------------------------------------*/
#include <stdio.h>
#include <float.h>
#include <math.h>
#include <string.h>
#include <cpl.h>
#include <muse.h>
#include "muse_skyflat_z.h"

/*---------------------------------------------------------------------------*
 *                              Functions code                               *
 *---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/**
  @brief    Add the QC parameters to the combined skyflat
  @param    aImage   the output master image
  @param    aList    the image list of input files
 */
/*---------------------------------------------------------------------------*/
static void
muse_skyflat_qc_header(muse_image *aImage, muse_imagelist *aList)
{
  cpl_msg_debug(__func__, "Adding QC keywords");

  unsigned stats = CPL_STATS_MEDIAN | CPL_STATS_MEAN | CPL_STATS_STDEV
                 | CPL_STATS_MIN | CPL_STATS_MAX;

  /* write the image statistics for input skyflat field exposures */
  unsigned int i;
  for (i = 0; i < muse_imagelist_get_size(aList); i++) {
    char *keyword = cpl_sprintf(QC_SKYFLAT_PREFIXi, i+1);
    muse_basicproc_stats_append_header(muse_imagelist_get(aList, i)->data,
                                       aImage->header, keyword, stats);
    cpl_free(keyword);
    keyword = cpl_sprintf(QC_SKYFLAT_PREFIXi" "QC_BASIC_NSATURATED, i+1);
    int nsaturated = cpl_propertylist_get_int(muse_imagelist_get(aList, i)->header,
                                              MUSE_HDR_TMP_NSAT);
    cpl_propertylist_update_int(aImage->header, keyword, nsaturated);
    cpl_free(keyword);
  } /* for i (all images in list) */

  /* properties of the resulting master frame */
  stats |= CPL_STATS_FLUX;
  muse_basicproc_stats_append_header(aImage->data, aImage->header,
                                     QC_SKYFLAT_MASTER_PREFIX, stats);
} /* muse_skyflat_qc_header() */

/*---------------------------------------------------------------------------*/
/**
  @brief    Add integrated fluxes as QC statistics to the master skyflat image.
  @param    aImage    the combined master skyflat, before normalization
  @param    aTrace    the trace table
  @param    aWave     the wavelength calibration table
  @param    aParams   recipe parameters, here lambdamin/-max are needed
  @return   The average value over all valid pixels, computed from the
            integrated flux or a negative value on error.
  @note This function also changes aImage in that it interpolates all bad pixels
        in the data component.

  @error{set CPL_ERROR_ILLEGAL_INPUT\, return -1,
         pixel table could not be created from the input data}
 */
/*---------------------------------------------------------------------------*/
static double
muse_skyflat_qc_integrate_flux(muse_image *aImage, cpl_table *aTrace,
                               cpl_table *aWave, muse_skyflat_params_t *aParams)
{
  cpl_msg_debug(__func__, "Adding flux statistics for QC, from %.1f to %.1f "
                "Angstrom", aParams->lambdamin, aParams->lambdamax);

  /* interpolate all bad pixels */
  muse_image_reject_from_dq(aImage);
  cpl_detector_interpolate_rejected(aImage->data);

  /* create pixel table from master sky flat */
  muse_pixtable *pt = muse_pixtable_create(aImage, aTrace, aWave, NULL);
  if (!pt) {
    cpl_error_set(__func__, CPL_ERROR_ILLEGAL_INPUT);
    return -1;
  }

  /* cut the pixel table to the requested wavelength range */
  muse_pixtable_restrict_wavelength(pt, aParams->lambdamin, aParams->lambdamax);
  /* then split up into slices */
  muse_pixtable **slicepts = muse_pixtable_extracted_get_slices(pt);
  /* integrate the flux over each slice, then sum up over all slices */
  int i, nint = 0,
      nslice = muse_pixtable_extracted_get_size(slicepts);
  double intflux = 0.;
  for (i = 0; i < nslice; i++) {
    double flux = cpl_table_get_column_mean(slicepts[i]->table,
                                            MUSE_PIXTABLE_DATA);
    int npix = muse_pixtable_get_nrow(slicepts[i]);
    flux *= npix; /* we need the sum */
    intflux += flux;
    nint += npix;

    char *keyword = cpl_sprintf(QC_SKYFLAT_MASTER_SLICEj_INTFLUX, i+1);
    cpl_propertylist_update_float(aImage->header, keyword, flux);
    cpl_free(keyword);
  } /* for i (all slices) */
  cpl_propertylist_update_float(aImage->header, QC_SKYFLAT_MASTER_INTFLUX, intflux);
  muse_pixtable_extracted_delete(slicepts);
  muse_pixtable_delete(pt);

  return intflux / nint;
} /* muse_skyflat_qc_integrate_flux() */

/*---------------------------------------------------------------------------*/
/**
  @brief    Interpret the command line options and execute the data processing
  @param    aProcessing   the processing structure
  @param    aParams       the parameters list
  @return   0 if everything is ok, -1 something went wrong
 */
/*---------------------------------------------------------------------------*/
int
muse_skyflat_compute(muse_processing *aProcessing,
                     muse_skyflat_params_t *aParams)
{
  muse_basicproc_params *bpars = muse_basicproc_params_new(aProcessing->parameters,
                                                           "muse.muse_skyflat");
  muse_imagelist *images = muse_basicproc_load(aProcessing, aParams->nifu, bpars);
  muse_basicproc_params_delete(bpars);
  cpl_ensure(images, cpl_error_get_code(), -1);

  muse_combinepar *cpars = muse_combinepar_new(aProcessing->parameters,
                                               "muse.muse_skyflat");
  muse_image *masterimage = muse_combine_images(cpars, images);
  muse_combinepar_delete(cpars);
  if (!masterimage) {
    cpl_msg_error(__func__, "Combining input frames failed!");
    muse_imagelist_delete(images);
    return -1;
  }
  cpl_propertylist_erase_regexp(masterimage->header, MUSE_WCS_KEYS, 0);

  muse_skyflat_qc_header(masterimage, images);
  muse_imagelist_delete(images);

  cpl_table *tracetable = muse_table_load(aProcessing, MUSE_TAG_TRACE_TABLE,
                                          aParams->nifu);
  cpl_table *wavecaltable = muse_table_load(aProcessing, MUSE_TAG_WAVECAL_TABLE,
                                            aParams->nifu);
  if (!tracetable || !wavecaltable) {
    cpl_msg_error(__func__, "Calibration could not be loaded:%s%s",
                  !tracetable ? " "MUSE_TAG_TRACE_TABLE : "",
                  !wavecaltable ? " "MUSE_TAG_WAVECAL_TABLE : "");
    /* try to clean up in case some files were successfully loaded */
    muse_image_delete(masterimage);
    cpl_table_delete(tracetable);
    cpl_table_delete(wavecaltable);
    return -1;
  }
  double mean = muse_skyflat_qc_integrate_flux(masterimage, tracetable,
                                               wavecaltable, aParams);
  cpl_table_delete(tracetable);
  cpl_table_delete(wavecaltable);
  if (mean < 0) {
    cpl_msg_error(__func__, "Flux integration failed: %s",
                  cpl_error_get_message());
    muse_image_delete(masterimage);
    return -1;
  }

  /* if requested, normalize the output skyflat-field to 1  *
   * using the mean value computed from the integrated flux */
  if (aParams->normalize) {
    muse_image_scale(masterimage, 1. / mean);
    /* EXPTIME keyword does not make sense any more for normalized skyflat */
    cpl_propertylist_erase(masterimage->header, "EXPTIME");
  }

  /* add saturation and slice image statistics as QC */
  muse_basicproc_qc_saturated(masterimage, QC_SKYFLAT_MASTER_PREFIX);
  int rc = muse_processing_save_image(aProcessing, aParams->nifu, masterimage,
                                      MUSE_TAG_MASTER_SKYFLAT);
  muse_image_delete(masterimage);
  return rc == CPL_ERROR_NONE ? 0 : -1;
} /* muse_skyflat_compute() */
