/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* vim:set sw=2 sts=2 et cin: */
/*
 *
 * This file is part of the MUSE Instrument Pipeline
 * Copyright (C) 2005-2011 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*----------------------------------------------------------------------------*
 *                              Includes                                      *
 *----------------------------------------------------------------------------*/
#include <stdio.h>
#include <float.h>
#include <math.h>
#include <string.h>
#include <cpl.h>
#include <muse.h>
#include "muse_wavecal_z.h"

/*---------------------------------------------------------------------------*
 *                             Functions code                                *
 *---------------------------------------------------------------------------*/
static cpl_error_code
muse_wavecal_save_lsf_params(cpl_propertylist *aHeader,
                             const muse_lsf_params **aSlicePars,
                             muse_processing *aProcessing, int aIFU)
{
  cpl_error_code r = CPL_ERROR_NONE;
  if (aSlicePars != NULL) {
    cpl_frame *frame
      = muse_processing_new_frame(aProcessing, aIFU, aHeader,
                                  "LSF_TABLE", CPL_FRAME_TYPE_TABLE);
    if (frame != NULL) {
      const char *filename = cpl_frame_get_filename(frame);
      cpl_msg_info(__func__, "Saving LSF table as %s", filename);
      char *channel = cpl_sprintf("CHAN%02d", aIFU);
      cpl_propertylist_update_string(aHeader, "EXTNAME", channel);
      cpl_free(channel);
      r = cpl_propertylist_save(aHeader, filename, CPL_IO_CREATE);
      r = muse_lsf_params_save(aSlicePars, filename);
      if (r == CPL_ERROR_NONE) {
        cpl_frameset_insert(aProcessing->outputFrames, frame);
      } else {
        cpl_frame_delete(frame);
      }
    }
  }
  return r;
} /* muse_wavecal_save_lsf_params() */

static muse_lsf_params *
muse_wavecal_compute_slice_lsf(cpl_table *aLines,
                               muse_pixtable *aPixtable,
                               muse_lsf_params *aFirstguess,
                               int max_iterations,
                               const muse_sky_fit_params *slice_fit_params)

{
  uint32_t origin = (uint32_t)cpl_table_get_int(aPixtable->table,
                                                MUSE_PIXTABLE_ORIGIN, 0, NULL);
  unsigned short i_ifu = muse_pixtable_origin_get_ifu(origin),
                 i_slice = muse_pixtable_origin_get_slice(origin);
  cpl_size nrows = muse_pixtable_get_nrow(aPixtable);

  cpl_msg_info(__func__, "processing slice %hu.%hu with %"CPL_SIZE_FORMAT" rows",
               i_ifu, i_slice, nrows);

  cpl_propertylist *order = cpl_propertylist_new();
  cpl_propertylist_append_bool(order, MUSE_PIXTABLE_LAMBDA, CPL_FALSE);
  cpl_table_sort(aPixtable->table, order);
  cpl_propertylist_delete(order);

  cpl_array *lambda = NULL;
  if (cpl_table_get_column_type(aPixtable->table, MUSE_PIXTABLE_LAMBDA)
      == CPL_TYPE_DOUBLE) {
    lambda = muse_cpltable_extract_column(aPixtable->table,
                                          MUSE_PIXTABLE_LAMBDA);
  } else {
    cpl_table_cast_column(aPixtable->table, MUSE_PIXTABLE_LAMBDA,
                          "lambda_double", CPL_TYPE_DOUBLE);
    lambda = muse_cpltable_extract_column(aPixtable->table, "lambda_double");
  }
  cpl_array *data = muse_cpltable_extract_column(aPixtable->table,
                                                 MUSE_PIXTABLE_DATA);
  cpl_array *stat = muse_cpltable_extract_column(aPixtable->table,
                                                 MUSE_PIXTABLE_STAT);

  muse_lsf_params *lsf
    = muse_lsf_params_fit(lambda, data, stat, aLines, aFirstguess,
                                   max_iterations, slice_fit_params);
  lsf->slice = i_slice;
  lsf->ifu = i_ifu;

  if (cpl_table_has_column(aPixtable->table, "lambda_double")) {
    cpl_table_erase_column(aPixtable->table, "lambda_double");
  }

  cpl_array_unwrap(lambda);
  cpl_array_unwrap(data);
  cpl_array_unwrap(stat);

  return lsf;
} /* muse_wavecal_compute_slice_lsf() */

static muse_lsf_params **
muse_wavecal_compute_lsf(cpl_table *aLines,
                         muse_pixtable *aPixtable,
                         muse_lsf_params **aFirstguess,
                         int max_iterations,
                         const muse_sky_fit_params *slice_fit_params)
{
  muse_pixtable **slice_pixtable = muse_pixtable_extracted_get_slices(aPixtable);
  int n_slices = muse_pixtable_extracted_get_size(slice_pixtable);
  int i_slice;

  muse_lsf_params **lsfParams 
    = cpl_calloc(n_slices+1, sizeof(muse_lsf_params *));
  #pragma omp parallel for default(none) num_threads(2)  /* as req. by Ralf */ \
          shared(aLines, n_slices, i_slice, lsfParams, slice_pixtable,         \
                 aFirstguess, max_iterations, slice_fit_params)
  for (i_slice = 0; i_slice < n_slices; i_slice++) {
    uint32_t origin
      = (uint32_t)cpl_table_get_int(slice_pixtable[i_slice]->table,
                                    MUSE_PIXTABLE_ORIGIN, 0, NULL);
    unsigned short ifu = muse_pixtable_origin_get_ifu(origin),
                   slice = muse_pixtable_origin_get_slice(origin);
    muse_lsf_params *firstguess_slice
      = muse_lsf_params_get(aFirstguess, ifu, slice);

    cpl_errorstate prestate = cpl_errorstate_get();
    lsfParams[i_slice]
      = muse_wavecal_compute_slice_lsf(aLines, slice_pixtable[i_slice],
                                       firstguess_slice, max_iterations,
                                       slice_fit_params);
    if (!cpl_errorstate_is_equal(prestate)) {
      cpl_msg_error(__func__, "While processing slice %hu.%hu:", ifu, slice);
      cpl_errorstate_dump(prestate, CPL_FALSE, NULL);
      cpl_errorstate_set(prestate);
    }
  }
  muse_pixtable_extracted_delete(slice_pixtable);
  return lsfParams;
} /* muse_wavecal_compute_lsf() */

/*----------------------------------------------------------------------------*/
/**
  @brief    Create a header with the QC parameters for the LSF parameters.
  @param    aMaster Computed sky parameters
  @return   Header with the QC parameters for the sky.
 */
/*----------------------------------------------------------------------------*/
static cpl_error_code
muse_wavecal_qc_lsf(cpl_propertylist *aHeader,
                    const muse_lsf_params **aSlicePars)
{
  double wavelength[] = { 5577.35, 8827.25, -1. };

  char keyword[KEYWORD_LENGTH];
  int i;
  for (i = 0; wavelength[i] > 0; i++) {
    snprintf(keyword, KEYWORD_LENGTH, "ESO QC LSF%i AWAV", i+1);
    cpl_propertylist_append_double(aHeader, keyword, wavelength[i]);
    const muse_lsf_params **slice;
    for (slice = aSlicePars; *slice != NULL; slice++) {
      double width
        = muse_cplarray_poly1d_double(wavelength[i] - (*slice)->lambda_ref,
                                      (*slice)->lsf_width);
      snprintf(keyword, KEYWORD_LENGTH,
               "ESO QC IFU%i SLICE%i LSF%i WIDTH",
               (*slice)->ifu, (*slice)->slice, i+1);
      cpl_propertylist_append_double(aHeader, keyword, width);
    }

    for (slice = aSlicePars; *slice != NULL; slice++) {
      double offset = (*slice)->offset +
        ((*slice)->refraction - 1) * wavelength[i];
      snprintf(keyword, KEYWORD_LENGTH,
               "ESO QC IFU%i SLICE%i LSF%i OFFSET",
               (*slice)->ifu, (*slice)->slice, i+1);
      cpl_propertylist_append_double(aHeader, keyword, offset);
    }
    for (slice = aSlicePars; *slice != NULL; slice++) {
      double fwhm = muse_lsf_fwhm_lambda(*slice, wavelength[i],
                                         1000., 0.01, 1000., NULL);
      snprintf(keyword, KEYWORD_LENGTH,
               "ESO QC IFU%i SLICE%i LSF%i FWHM",
               (*slice)->ifu, (*slice)->slice, i+1);
      cpl_propertylist_append_float(aHeader, keyword, fwhm);
    }
  }
  const muse_lsf_params **slice;
  for (slice = aSlicePars; *slice != NULL; slice++) {
    double wcal_error = (*slice)->refraction - 1;
    snprintf(keyword, KEYWORD_LENGTH,
             "ESO QC IFU%i SLICE%i WAVECAL ERROR",
             (*slice)->ifu, (*slice)->slice);
    cpl_propertylist_append_double(aHeader, keyword, wcal_error);
  }

  return CPL_ERROR_NONE;
} /* muse_wavecal_qc_lsf() */

/*---------------------------------------------------------------------------*/
/**
  @brief    Sum the image list and save the resulting image.
  @param    aImages       the image list to combine
  @param    aProcessing   the processing structure
  @param    aIFU          the IFU/channel number
  @return   the combined image or NULL on error

  Take care that all headers are transferred / erased before saving the result.
 */
/*---------------------------------------------------------------------------*/
static muse_image *
muse_wavecal_sum_and_save(muse_imagelist *aImages, muse_processing *aProcessing,
                          int aIFU)
{
  cpl_ensure(aImages && aProcessing, CPL_ERROR_NULL_INPUT, NULL);

  /* do the final (sum!) image combination if requested */
  char *pname = cpl_sprintf("muse.%s.combine", aProcessing->recipeName);
  cpl_parameter *param = cpl_parameterlist_find(aProcessing->parameters, pname);
  char *porig = cpl_strdup(cpl_parameter_get_string(param));
  cpl_parameter_set_string(param, "sum");
  cpl_free(pname);
  muse_combinepar *cpars = muse_combinepar_new(aProcessing->parameters,
                                               "muse.muse_wavecal");
  cpl_parameter_set_string(param, porig);
  cpl_free(porig);
  muse_image *image = muse_combine_images(cpars, aImages);
  muse_combinepar_delete(cpars);
  if (!image) {
    cpl_error_set_message(__func__, CPL_ERROR_ILLEGAL_OUTPUT,
                          "summing per-lamp images did not work");
    return NULL;
  }

  /* transfer NSATURATION headers from lamp-combined images to final combined image */
  /* loop over lamps (entries in the imagelist) */
  unsigned int k;
  for (k = 0; k < muse_imagelist_get_size(aImages); k++) {
    cpl_errorstate state = cpl_errorstate_get();
    unsigned int i = 1;
    while (cpl_errorstate_is_equal(state)) {
      char *kwget = cpl_sprintf(QC_WAVECAL_PREFIXi" "QC_BASIC_NSATURATED, i),
           *kwout = cpl_sprintf(QC_WAVECAL_PREFIXli" "QC_BASIC_NSATURATED, k+1, i);
      /* this will change the errorstate if the keyword doesn't exist: */
      int nsaturated = cpl_propertylist_get_int(muse_imagelist_get(aImages, k)->header,
                                                kwget);
      if (cpl_errorstate_is_equal(state)) {
        cpl_propertylist_update_int(image->header, kwout, nsaturated);
      }
      cpl_free(kwget);
      cpl_free(kwout);
      i++;
    } /* while (errorstate is the same) */
    /* reset the old errorstate, we don't want to leak *
     * it outside, since it's not a real error         */
    cpl_errorstate_set(state);
  } /* for k (all images in list) */
  /* count saturated pixels in final combined image */
  muse_basicproc_qc_saturated(image, QC_WAVECAL_PREFIX);

  /* save the pre-processed image to file                                *
   * (this should be done, even if the wavelength calibration fails so   *
   * that the user can check what input files were used for the wavecal) */
  muse_processing_save_image(aProcessing, aIFU, image,
                             MUSE_TAG_ARC_RED);

  return image;
} /* muse_wavecal_sum_and_save() */

/*---------------------------------------------------------------------------*/
/**
  @brief    Interpret the command line options and execute the data processing
  @param    aProcessing   the processing structure
  @param    aParams       the parameters list
  @return   0 if everything is ok, -1 something went wrong
 */
/*---------------------------------------------------------------------------*/
int
muse_wavecal_compute(muse_processing *aProcessing,
                     muse_wavecal_params_t *aParams)
{
  muse_wave_weighting_type fweight = MUSE_WAVE_WEIGHTING_UNIFORM;
  if (aParams->fitweighting == MUSE_WAVECAL_PARAM_FITWEIGHTING_CERR) {
    fweight = MUSE_WAVE_WEIGHTING_CERR;
  } else if (aParams->fitweighting == MUSE_WAVECAL_PARAM_FITWEIGHTING_SCATTER) {
    fweight = MUSE_WAVE_WEIGHTING_SCATTER;
  } else if (aParams->fitweighting == MUSE_WAVECAL_PARAM_FITWEIGHTING_CERRSCATTER) {
    fweight = MUSE_WAVE_WEIGHTING_CERRSCATTER;
  } else if (aParams->fitweighting != MUSE_WAVECAL_PARAM_FITWEIGHTING_UNIFORM) {
    cpl_msg_error(__func__, "unknown fitweighting method \"%s\"",
                  aParams->fitweighting_s);
    return -1;
  }

  muse_basicproc_params *bpars = muse_basicproc_params_new(aProcessing->parameters,
                                                           "muse.muse_wavecal");
  cpl_frameset **labeledframes = NULL;
  muse_imagelist *images = muse_basicproc_combine_images_lampwise(aProcessing,
                                                                  aParams->nifu,
                                                                  bpars,
                                                                  &labeledframes);
  muse_basicproc_params_delete(bpars);
  cpl_ensure(images, cpl_error_get_code(), -1);
  /* save all resulting images, but only if there were different lamp *
   * setups; otherwise we only save the combined image, see below     */
  cpl_frameset *used = aProcessing->usedFrames;
  unsigned int k, kimages = muse_imagelist_get_size(images);
  cpl_boolean needsum = !aParams->lampwise || aParams->resample || aParams->lsf;
  for (k = 0; k < kimages; k++) {
    /* only save individual files if different from combined one */
    if (!(kimages == 1 && needsum)) {
      muse_image *image = muse_imagelist_get(images, k);
      if (labeledframes) { /* replace usedFrames with ones used for _this_ image */
        aProcessing->usedFrames = labeledframes[k];
      }
      muse_basicproc_qc_saturated(image, QC_WAVECAL_PREFIX);
      cpl_propertylist_erase_regexp(image->header, MUSE_WCS_KEYS, 0);
      muse_processing_save_image(aProcessing, aParams->nifu, image,
                                 MUSE_TAG_ARC_RED_LAMP);
    }
    /* delete the corresponding frameset in any case, if present */
    if (labeledframes) {
      cpl_frameset_delete(labeledframes[k]);
    }
  } /* for k (all images in list) */
  cpl_free(labeledframes);
  aProcessing->usedFrames = used;
  muse_image *masterimage = NULL;
  if (needsum) {
    masterimage = muse_wavecal_sum_and_save(images, aProcessing, aParams->nifu);
  } /* if !lampwise */

  /* line catalog is needed in any case */
  cpl_table *linelist = muse_table_load(aProcessing, MUSE_TAG_LINE_CATALOG, 0);
  cpl_propertylist *linehead = muse_propertylist_load(aProcessing,
                                                      MUSE_TAG_LINE_CATALOG);
  int rc = muse_wave_lines_check(linelist, linehead) ? 0 : -1;
  cpl_propertylist_delete(linehead);
  if (rc) {
    cpl_table_delete(linelist);
    muse_image_delete(masterimage);
    muse_imagelist_delete(images);
    return rc;
  }
  /* trace table is needed in any case if we made it here */
  cpl_table *tracetable = muse_table_load(aProcessing, MUSE_TAG_TRACE_TABLE,
                                          aParams->nifu);

  /* the main step: compute the wavelength solution, add QC parameters   *
   * to the FITS headers of the input exposure(s), since the output      *
   * table doesn't carry headers, and create an output header for saving *
   * by copying most of the input headers.                               */
  cpl_table *wavecaltable = NULL;
  cpl_propertylist *header = NULL;
  muse_wave_params *p = muse_wave_params_new();
  p->xorder = aParams->xorder;
  p->yorder = aParams->yorder;
  p->detsigma = aParams->sigma;
  p->ddisp = aParams->dres;
  p->tolerance = aParams->tolerance;
  p->linesigma = aParams->linesigma;
  p->rflag = aParams->residuals; /* the residuals component itself stays NULL */
  p->fitsigma = aParams->fitsigma;
  p->fitweighting = fweight;
  if (aParams->lampwise) { /* separate image handling (new method) */
    wavecaltable = muse_wave_calib_lampwise(images, tracetable, linelist, p);
    header = cpl_propertylist_duplicate(muse_imagelist_get(images, 0)->header);
  } else { /* direct measurement and fit on single combined image (old method) */
    wavecaltable = muse_wave_calib(masterimage, tracetable, linelist, p);
    header = cpl_propertylist_duplicate(masterimage->header);
  }
  if (!wavecaltable) { /* failure of the main step, return now */
    muse_image_delete(masterimage);
    muse_imagelist_delete(images);
    cpl_table_delete(tracetable);
    cpl_table_delete(linelist);
    cpl_propertylist_delete(header);
    muse_wave_params_delete(p);
    return -1;
  }

  /* create table header by copying most of the image header *
   * and save the wavelength solution in this header         */
  cpl_propertylist_erase_regexp(header,
                                "^SIMPLE$|^BITPIX$|^NAXIS|^EXTEND$|^XTENSION$|"
                                "^DATASUM$|^DATAMIN$|^DATAMAX$|^DATAMD5$|"
                                "^PCOUNT$|^GCOUNT$|^HDUVERS$|^BLANK$|"
                                "^BZERO$|^BSCALE$|^BUNIT$|^CHECKSUM$|^INHERIT$|"
                                "^PIPEFILE$|^ESO PRO ", 0);
  muse_processing_save_table(aProcessing, aParams->nifu, wavecaltable, header,
                             MUSE_TAG_WAVECAL_TABLE, MUSE_TABLE_TYPE_CPL);
  if (p->residuals) {
    /* don't want to save QC headers with the residuals table: */
    cpl_propertylist_erase_regexp(header, QC_WAVECAL_PREFIX, 0);
    muse_processing_save_table(aProcessing, aParams->nifu, p->residuals, header,
                               MUSE_TAG_WAVECAL_DEBUG, MUSE_TABLE_TYPE_CPL);
  }
  muse_wave_params_delete(p);
  cpl_propertylist_delete(header);

  /* now the extra steps: resampling, LSF determination, or wavemap creation */
  muse_pixtable *pixtable = NULL;
  if (aParams->resample || aParams->lsf) {
    /* to visualize wavelength calibration, or for *
     * LSF fitting, create the pixel table now */
    pixtable = muse_pixtable_create(masterimage, tracetable, wavecaltable, NULL);
    if (!pixtable) {
      cpl_error_set_message(__func__, CPL_ERROR_ILLEGAL_INPUT, "pixel table "
                            "creation failed!");
      if (!getenv("MUSE_EXPERT_USER")) {
        muse_image_delete(masterimage);
        cpl_table_delete(wavecaltable);
        cpl_table_delete(tracetable);
        muse_imagelist_delete(images);
        cpl_table_delete(linelist);
        return -1;
      } /* if non-expert user */
    } /* if no pixel table */
  } /* if pixel table needed */

  if (aParams->resample) {
    /* to visualize wavelength calibration, use the     *
     * pixel table to resample the data onto a 2D image */
    muse_image *resampled = muse_resampling_image(pixtable,
                                                  MUSE_RESAMPLE_WEIGHTED_RENKA,
                                                  1.0, 1.25);
    if (resampled) {
      /* don't want to save QC headers with the resampled image: */
      cpl_propertylist_erase_regexp(resampled->header, QC_WAVECAL_PREFIX, 0);
      muse_processing_save_cimage(aProcessing, aParams->nifu, resampled->data,
                                  resampled->header, MUSE_TAG_ARC_RESAMP);
      muse_image_delete(resampled);
    }
  } /* if resample */

  if (aParams->wavemap) {
    muse_image *image = masterimage ? masterimage
                                    : muse_imagelist_get(images, 0);
    cpl_image *map = muse_wave_map(image, wavecaltable, tracetable);
    /* simply use the header of the incoming image */
    muse_processing_save_cimage(aProcessing, aParams->nifu, map, image->header,
                                MUSE_TAG_WAVE_MAP);
    cpl_image_delete(map);
  } /* if wavemap */

  if (aParams->lsf && pixtable) {
    if (aParams->lsf_residuals) {
     /* Duplicate to have access to original data after fit for residual calc */
      cpl_table_duplicate_column(pixtable->table, "orig_data", pixtable->table,
                                 "data"); 
    }

    /* now we can throw away the original flux column and rename the one with *
     * the measurements to "flux" as needed by the LSF fitting routines       */
    cpl_table_erase_column(linelist, MUSE_LINE_CATALOG_FLUX);
    cpl_table_name_column(linelist, "Flux_for_LSF", "flux");
    /* erase unused (flux == 0.) and non-isolated (quality < 3) lines */
    cpl_table_unselect_all(linelist);
    cpl_table_or_selected_int(linelist, MUSE_LINE_CATALOG_QUALITY,
                              CPL_LESS_THAN, 3);
    cpl_table_or_selected_double(linelist, "flux", CPL_NOT_GREATER_THAN, 0.);
    cpl_table_erase_selected(linelist);
    /* sort list by decreasing flux and cut it to the 40 brightest lines */
    cpl_propertylist *flux_sort = cpl_propertylist_new();
    cpl_propertylist_append_bool(flux_sort, "flux", CPL_TRUE);
    cpl_table_sort(linelist, flux_sort);
    cpl_propertylist_delete(flux_sort);
    cpl_size nrows = 40;
    if (nrows < cpl_table_get_nrow(linelist)) {
      cpl_table_erase_window(linelist, nrows, cpl_table_get_nrow(linelist));
    }

    muse_sky_fit_params *slice_fit_params = muse_sky_fit_params_new
      (
       0, // aParams->slice_fit_offset,
       0, // aParams->slice_fit_refraction,
       0, // sensitivity is not used here
       1, // aParams->slice_fit_slit_width,
       1, // aParams->slice_fit_bin_width,
       3, // aParams->slice_fit_lsf_width + 1,
       1, //1, // aParams->slice_fit_h3 + 1,
       2, //2, // aParams->slice_fit_h4 + 1,
       1, //1, // aParams->slice_fit_h5 + 1,
       2  //2  // aParams->slice_fit_h6 + 1
       );

    muse_lsf_params **firstguess
      = muse_processing_lsf_params_load(aProcessing, aParams->nifu);
    muse_lsf_params **lsf
      = muse_wavecal_compute_lsf(linelist, pixtable, firstguess, 40,
                                 slice_fit_params);

    muse_lsf_params_delete(firstguess);
    muse_sky_fit_params_delete(slice_fit_params);

    cpl_propertylist *qc_header = cpl_propertylist_new();
    muse_wavecal_qc_lsf(qc_header, (const muse_lsf_params **)lsf);
    muse_wavecal_save_lsf_params(qc_header, (const muse_lsf_params **)lsf,
                                 aProcessing, aParams->nifu);
    muse_lsf_params_delete(lsf);
    cpl_propertylist_delete(qc_header);

    if (aParams->lsf_residuals) {
      muse_processing_save_table(aProcessing, aParams->nifu, pixtable, NULL,
                                 "LSF_RESIDUALS", MUSE_TABLE_TYPE_PIXTABLE);
    }
  } /* if lsf */

  /* clean up */
  muse_pixtable_delete(pixtable);
  cpl_table_delete(wavecaltable);
  muse_image_delete(masterimage);
  muse_imagelist_delete(images);
  cpl_table_delete(tracetable);
  cpl_table_delete(linelist);

  return rc; /* can only be 0 or -1 */
} /* muse_wavecal_compute() */
