/* -*- Mode: C; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 8 -*- */
/* vim:set sw=8 sts=8 et cin: */
/*
 * This file is part of the MUSE Instrument Pipeline
 * Copyright (C) 2007-2012 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

/*----------------------------------------------------------------------------* 
 *                               Includes                                     * 
 *----------------------------------------------------------------------------*/

#include <string.h>
#include <cpl.h>
#include <muclipm_make_image.h>

/*----------------------------------------------------------------------------*/
/**
 * @defgroup tools_musemakeima              Tool muse_makeima
 *
 * @brief   <b>muse_makeima</b>: standalone quick image reconstruction
 *
 * Written by Arlette Pecontal.
 */
/*----------------------------------------------------------------------------*/
int main(int argc, char **argv) 
{
	/* local variables */
	cpl_image *raw_img, *bias = NULL, *mask;
	cpl_propertylist *mask_hdr, *hdr;
	const char *idstring = "muse_makeima_main";
        char filename[256], extname[72], *pt_file;
	cpl_image *computed_img, *tmp_img;
	long xstart[MUCLIPM_NB_SLICES*MUCLIPM_NB_SUBSLICERS];
	long xend[MUCLIPM_NB_SLICES*MUCLIPM_NB_SUBSLICERS];
	char property[72];
	int nbchannel, nifu;
	long raw_img_nx, raw_img_ny, i;
	long mask_nx, mask_ny;
	long bias_nx, bias_ny;
	int n_IFU, nbext, ext[24], mask_nbext, bias_nbext, current_chan;
	float cpl_version;

	cpl_init(0);

        cpl_error_reset();
	cpl_msg_set_component_on();
/*         cpl_msg_set_level(CPL_MSG_DEBUG);  */

        if (argc < 3 || argc > 5) {
		cpl_msg_error(idstring, "Usage is: makeima science mask [bias] [flat]");

		cpl_end();
		return -1;
        }
	cpl_version = cpl_version_get_major() + 0.1*cpl_version_get_minor();
	printf("Running CPL VERSION %2.1f\n",cpl_version);

	/* create the output image */

        pt_file = strrchr(argv[1],'/');
	if (pt_file != NULL) {
        	sprintf(filename,"reconst_%s",pt_file+1);
		printf("Filename is %s\n",filename);
	} 
	else
	{
        	sprintf(filename,"reconst_%s",argv[1]);
	}

	/* eval nb of extensions in files */

	n_IFU = cpl_fits_count_extensions(argv[1]);

	if (n_IFU > 24) /* INM may put more extensions like COSMIC table ... */
		n_IFU = 24;

	printf("%d extensions will be processed\n", n_IFU);
	mask_nbext = cpl_fits_count_extensions(argv[2]);
	if (mask_nbext < n_IFU) {
		n_IFU = mask_nbext;
	}

	if (argc > 3) {
		bias_nbext = cpl_fits_count_extensions(argv[3]);

		if (bias_nbext < n_IFU) 
			n_IFU = bias_nbext;
	}

	/* find valid extensions into Science file */

	for (nbext = 1; nbext <= n_IFU; nbext++) { 
               	raw_img = cpl_image_load(argv[1], CPL_TYPE_FLOAT, 0, nbext); 
		if (!raw_img)
			ext[nbext-1] = 0 ;
		else {
			ext[nbext-1] = 1 ;
        		cpl_image_delete(raw_img);
		}
	}

	computed_img = cpl_image_new(MUCLIPM_NB_SUBSLICERS*MUCLIPM_NB_SPEC_PER_SLICE,
			MUCLIPM_NB_SLICES*n_IFU, CPL_TYPE_FLOAT);

	/* printf("TUNED for INM DATA !!!! Reverse order for channels\n"); */

	for (nbchannel = 0; nbchannel < n_IFU; nbchannel++) { 
	
		current_chan = nbchannel+1;

		if (!ext[nbchannel]) continue;
			
		tmp_img = cpl_image_new(MUCLIPM_NB_SUBSLICERS*MUCLIPM_NB_SPEC_PER_SLICE,
			MUCLIPM_NB_SLICES, CPL_TYPE_FLOAT);

                raw_img = cpl_image_load(argv[1], CPL_TYPE_FLOAT, 0, current_chan); 
	        if (!raw_img) {
                        cpl_msg_info(idstring,"%d extension(s) processed",current_chan);
		        break;
  	        }
		raw_img_nx = cpl_image_get_size_x(raw_img);
		raw_img_ny = cpl_image_get_size_y(raw_img);

		if (raw_img_nx == 0 || raw_img_ny == 0) continue;

		/* get channel number */
		hdr = cpl_propertylist_load(argv[1], current_chan);
		strcpy(extname, cpl_propertylist_get_string(hdr, "EXTNAME"));
		sscanf(extname, "CHAN%02d", &nifu);
		printf("Processing CHAN%d (extension %d)\n",nifu, nbchannel+1);
		/* to be disable for old INM data like Starfield  */
		
		if (n_IFU == 24)
			nifu = 24 - nifu;  
		
		/* nifu--; */
	        cpl_propertylist_delete(hdr);

		current_chan = cpl_fits_find_extension(argv[2],extname);
		if (current_chan < 0) continue;

		mask = cpl_image_load(argv[2], CPL_TYPE_INT, 0, current_chan); 
		if (!mask) {
			cpl_msg_error(idstring, "Could not load frame %s for channel %d: %s",
                  	argv[2], current_chan,cpl_error_get_message());
			continue;
  		}
		mask_nx = cpl_image_get_size_x(mask);
		mask_ny = cpl_image_get_size_y(mask);

		if (mask_nx != raw_img_nx || mask_ny != raw_img_ny) {
			cpl_msg_error(idstring, "Science image and mask don't have the same size");
			cpl_end();
			return -1 ;
		}

		mask_hdr = cpl_propertylist_load(argv[2], current_chan);
		if (!mask_hdr) {
			cpl_propertylist_delete(mask_hdr);
	       		cpl_image_delete(mask);
			continue;
		}

		/* read slice limitder */

       		for (i=0; i< MUCLIPM_NB_SLICES*MUCLIPM_NB_SUBSLICERS; i++) {
	               sprintf(property,"ESO DET SLICE%ld XSTART",i+1);
	               xstart[i] = cpl_propertylist_get_int(mask_hdr, property);
	               sprintf(property,"ESO DET SLICE%ld XEND",i+1);
	               xend[i] = cpl_propertylist_get_int(mask_hdr, property);
		       /* printf("Slice %ld: limits [%ld,%ld]\n", i+1, xstart[i],xend[i]); */
		       if (xstart[i] >= xend[i]) {
			       cpl_msg_error(idstring, "Channel %d: inconsistant slice limits in X [%ld,%ld] for slice %ld",
					                               current_chan,xstart[i],xend[i], i+1);
		               return -1; 
		       }
	        }
		
       		for (i=1; i< MUCLIPM_NB_SLICES*MUCLIPM_NB_SUBSLICERS; i++) {
			
		       if (xstart[i] <= xend[i-1]) {
			       cpl_msg_error(idstring, "Channel %d: inconsistant interslice for slices %ld and %ld (%ld,%ld)",
					                               current_chan,i-1, i, xend[i-1],xstart[i]);
		               return -1; 
		       }
	        }

		if (argc > 3) {
			bias = cpl_image_load(argv[3], CPL_TYPE_FLOAT, 0, current_chan); 
			if (!bias) {
				cpl_msg_error(idstring, "Could not load frame %s for channel %d: %s",
                  		argv[3], current_chan, cpl_error_get_message());
				continue;
  			}
			bias_nx = cpl_image_get_size_x(bias);
			bias_ny = cpl_image_get_size_y(bias);

			if (bias_nx != raw_img_nx || bias_ny != raw_img_ny) {
				cpl_msg_error(idstring, "Science image and bias have not the same size");
				cpl_end();
				return -1 ;
			}
		}
		else {
			bias = NULL;
                }

		/* compute reconstructed image */
		if ( muclipm_make_image(raw_img, bias, NULL, mask, xstart, xend, &tmp_img) != CPL_ERROR_NONE) 
		{
			cpl_msg_info(idstring,"Error while computing reconstructed image");
			cpl_end();
			return -1;
		};

		if (n_IFU < 24) 
		{
			if (cpl_image_copy(computed_img,tmp_img,1,nbchannel*MUCLIPM_NB_SLICES+1)) {
				cpl_msg_error(idstring, "Could not copy subframe for channel %d: %s",
                  			nifu+1, cpl_error_get_message());
				cpl_end();
			}
		}
		else 
		{
			if (cpl_image_copy(computed_img,tmp_img,1,nifu*MUCLIPM_NB_SLICES+1)) {
				cpl_msg_error(idstring, "Could not copy subframe for channel %d: %s",
                  			nifu+1, cpl_error_get_message());
				cpl_end();
			}
		}
	
	        /* free memory associated to input image */
	        cpl_image_delete(raw_img);

		/* test first if bias was used */ 
		if (bias != NULL)
			cpl_image_delete(bias);

		cpl_image_delete(mask);
	        cpl_propertylist_delete(mask_hdr);

		cpl_image_delete(tmp_img);
	}

        hdr = cpl_propertylist_load(pt_file, 0); 

        cpl_image_flip(computed_img, 0);
        cpl_image_save(computed_img, filename, CPL_BPP_IEEE_FLOAT, hdr, CPL_IO_DEFAULT);
	cpl_image_delete(computed_img);

	cpl_end();
	return 0;
}
